import functools
import sys
import typing as ta

from ...... import check
from ...... import lang
from .base import AsyncsBackend


if ta.TYPE_CHECKING:
    import asyncio
else:
    asyncio = lang.proxy_import('asyncio')


##


class AsyncioAsyncsBackend(AsyncsBackend):
    name = 'asyncio'

    def is_available(self) -> bool:
        return True

    def is_imported(self) -> bool:
        return 'asyncio' in sys.modules

    #

    def wrap_runner(self, fn):
        @functools.wraps(fn)
        def wrapper(**kwargs):
            with asyncio.Runner(loop_factory=asyncio.get_event_loop_policy().new_event_loop) as runner:
                loop_cls = type(runner.get_loop())
                check.equal(loop_cls.__module__.split('.')[0], 'asyncio')
                return runner.run(fn(**kwargs))

        return wrapper



"""
TODO:
 - replace with codecs.StreamReader
"""
import io
import typing as ta


##


class DelimitedEscaping:
    def __init__(
            self,
            delimit_char: str,
            quote_char: str,
            escape_char: str,
            escaped_chars: ta.Iterable[str] = (),
    ) -> None:
        super().__init__()

        self._delimit_char = delimit_char
        self._quote_char = quote_char
        self._escape_char = escape_char
        self._escaped_chars = frozenset(escaped_chars)

        for c in [delimit_char, quote_char, escape_char]:
            if not isinstance(c, str) or len(c) != 1:
                raise TypeError(c)
        for c in self._escaped_chars:
            if not isinstance(c, str):
                raise TypeError(c)

        self._all_escaped_chars = frozenset({delimit_char, quote_char, escape_char} | self._escaped_chars)

    @property
    def delimit_char(self) -> str:
        return self._delimit_char

    def quote_char(self) -> str:
        return self._quote_char

    def escape_char(self) -> str:
        return self._escape_char

    def escaped_chars(self) -> frozenset[str]:
        return self._escaped_chars

    def all_escaped_chars(self) -> frozenset[str]:
        return self._all_escaped_chars

    def is_control_char(self, c: str) -> bool:
        if not len(c) == 1:
            raise TypeError(c)
        return c == self._delimit_char or c == self._quote_char or c == self._escape_char  # noqa

    def contains_escaped_char(self, s: str) -> bool:
        return any(c in self._all_escaped_chars for c in s)

    def escape(self, s: str) -> str:
        buf = io.StringIO()
        for c in s:
            if c in self._all_escaped_chars:
                buf.write(self._escape_char)
            buf.write(c)
        return buf.getvalue()

    def unescape(self, s: str) -> str:
        buf = io.StringIO()
        i = 0
        while i < len(s):
            c = s[i]
            if c == self._escape_char:
                if i > (len(s) - 2):
                    raise ValueError(s)
                i += 1
                buf.write(s[i])
            else:
                if c in self._all_escaped_chars:
                    raise ValueError(s)
                buf.write(c)
            i += 1
        return buf.getvalue()

    def quote(self, s: str) -> str:
        if self.contains_escaped_char(s):
            return self._quote_char + self.escape(s) + self._quote_char
        else:
            return s

    def unquote(self, s: str) -> str:
        if s and s[0] == self._quote_char:
            if len(s) < 2 or s[-1] != self._quote_char:
                raise ValueError(s)
            return self.unescape(s[1:-1])
        else:
            return s

    def delimit_many(self, strs: ta.Iterable[str]) -> str:
        if isinstance(strs, str):
            raise TypeError(strs)
        buf = io.StringIO()
        for count, s in enumerate(strs):
            if count:
                buf.write(self._delimit_char)
            if self.contains_escaped_char(s):
                buf.write(self.quote(s))
            else:
                buf.write(s)
        return buf.getvalue()

    def delimit(self, s: str) -> str:
        if not isinstance(s, str):
            raise TypeError(s)
        return self.delimit_many([s])

    def undelimit(self, s: str) -> list[str]:
        ret = []
        buf = io.StringIO()
        count = 0
        i = 0

        while i < len(s):
            c = s[i]

            if count:
                if c != self._delimit_char or i >= (len(s) - 1):
                    raise ValueError(s)
                i += 1
                c = s[i]

            quoted = c == self._quote_char
            if quoted:
                if i >= (len(s) - 1):
                    raise ValueError(s)
                i += 1
                c = s[i]
            unquoted = False

            while True:
                if c == self._delimit_char:
                    if not quoted:
                        break
                    buf.write(c)
                elif c == self._quote_char:
                    if not quoted:
                        raise ValueError(s)
                    unquoted = True
                    i += 1
                    break
                elif c == self._escape_char:
                    if not quoted or i > (len(s) - 2):
                        raise ValueError(s)
                    i += 1
                    buf.write(s[i])
                else:
                    if c in self._escaped_chars:
                        raise ValueError(s)
                    buf.write(c)

                i += 1
                if i == len(s):
                    break
                c = s[i]

            if quoted and not unquoted:
                raise ValueError(s)

            ret.append(buf.getvalue())
            buf.seek(0)
            buf.truncate()
            count += 1

        return ret



# ruff: noqa: UP007 UP045
# @omlish-lite
import datetime
import re
import typing as ta


##


def to_seconds(value: datetime.timedelta) -> float:
    return 86400 * value.days + value.seconds + 0.000001 * value.microseconds


def months_ago(date: datetime.date, num: int) -> datetime.date:
    ago_year = date.year
    ago_month = date.month - num
    while ago_month < 1:
        ago_year -= 1
        ago_month += 12
    while ago_month > 12:
        ago_year += 1
        ago_month -= 12
    return datetime.date(ago_year, ago_month, 1)


def parse_date(s: str, tz: ta.Optional[datetime.timezone] = None) -> datetime.date:
    if s.lower() in ['today', 'now']:
        return datetime.datetime.now(tz=tz)
    elif s.lower() == 'yesterday':
        return datetime.datetime.now(tz=tz) - datetime.timedelta(days=1)
    elif s.lower().endswith(' days ago'):
        num = int(s.split(' ', 1)[0])
        return datetime.datetime.now(tz=tz) - datetime.timedelta(days=num)
    elif s.lower().endswith(' months ago'):
        months = int(s.split(' ', 1)[0])
        return months_ago(datetime.datetime.now(tz=tz), months)
    else:
        return datetime.date(*map(int, s.split('-', 3)))


_TIMEDELTA_STR_RE = re.compile(
    r'^\s*'
    r'((?P<days>-?\d+)\s*days?,\s*)?'
    r'(?P<hours>\d?\d):(?P<minutes>\d\d)'
    r':(?P<seconds>\d\d+(\.\d+)?)'
    r'\s*$',
)


_TIMEDELTA_DHMS_RE = re.compile(
    r'^\s*'
    r'(?P<negative>-)?'
    r'((?P<days>\d+(\.\d+)?)\s*(d|days?))?'
    r',?\s*((?P<hours>\d+(\.\d+)?)\s*(h|hours?))?'
    r',?\s*((?P<minutes>\d+(\.\d+)?)\s*(m|minutes?))?'
    r',?\s*((?P<seconds>\d+(\.\d+)?)\s*(s|secs?|seconds?))?'
    r'\s*$',
)


def parse_timedelta(s: str) -> datetime.timedelta:
    match = _TIMEDELTA_DHMS_RE.match(s)
    if not match:
        match = _TIMEDELTA_STR_RE.match(s)
    if not match:
        raise ValueError
    timedelta_kwargs = {
        k: float(v)
        for k, v in match.groupdict().items()
        if k != 'negative' and v is not None
    }
    if not timedelta_kwargs:
        raise ValueError
    sign = -1 if match.groupdict().get('negative') else 1
    return sign * datetime.timedelta(**timedelta_kwargs)



import typing as ta

from .consts import MAX_LEN_CLIENT_TAG_DATA
from .consts import MAX_LEN_SERVER_TAG_DATA
from .consts import MAX_LEN_TAGS_FROM_CLIENT
from .errors import BadCharactersError
from .errors import BadParamError
from .errors import CommandMissingError
from .errors import InvalidTagContentError
from .errors import TagsTooLongError
from .message import Message
from .tags import escape_tag_value
from .tags import validate_tag_name
from .tags import validate_tag_value
from .utils import find_utf8_truncation_point


##


def param_requires_trailing(param: str) -> bool:
    return len(param) == 0 or ' ' in param or param[0] == ':'


class RenderedLine(ta.NamedTuple):
    raw: bytes

    truncated: bool = False


def render_line_(
        msg: Message,
        *,
        tag_limit: int | None = None,
        client_only_tag_data_limit: int | None = None,
        server_added_tag_data_limit: int | None = None,
        truncate_len: int | None = None,
) -> RenderedLine:
    if not msg.command:
        raise CommandMissingError

    buf = bytearray()
    len_regular_tags = len_client_only_tags = 0

    # Write the tags, computing the budgets for client-only tags and regular tags
    if msg.tags or msg.client_only_tags:
        buf.append(ord('@'))
        first_tag = True
        tag_error = None

        def write_tags(tags: ta.Mapping[str, str]) -> None:
            nonlocal first_tag, tag_error
            for tag, val in tags.items():
                if not (validate_tag_name(tag) and validate_tag_value(val)):
                    tag_error = InvalidTagContentError
                if not first_tag:
                    buf.append(ord(';'))
                buf.extend(tag.encode('utf-8'))
                if val:
                    buf.append(ord('='))
                    buf.extend(escape_tag_value(val).encode('utf-8'))
                first_tag = False

        write_tags(msg.tags or {})
        len_regular_tags = len(buf) - 1
        write_tags(msg.client_only_tags or {})
        len_client_only_tags = (len(buf) - 1) - len_regular_tags
        if len_regular_tags:
            # Semicolon between regular and client-only tags is not counted
            len_client_only_tags -= 1
        buf.append(ord(' '))
        if tag_error:
            raise tag_error

    len_tags = len(buf)
    if tag_limit is not None and len(buf) > tag_limit:
        raise TagsTooLongError
    if (
            (client_only_tag_data_limit is not None and len_client_only_tags > client_only_tag_data_limit) or
            (server_added_tag_data_limit is not None and len_regular_tags > server_added_tag_data_limit)
    ):
        raise TagsTooLongError

    if msg.source:
        buf.append(ord(':'))
        buf.extend(msg.source.encode('utf-8'))
        buf.append(ord(' '))

    buf.extend(msg.command.encode('utf-8'))

    for i, param in enumerate(msg.params):
        buf.append(ord(' '))
        requires_trailing = param_requires_trailing(param)
        last_param = i == len(msg.params) - 1

        if (requires_trailing or msg.force_trailing) and last_param:
            buf.append(ord(':'))
        elif requires_trailing and not last_param:
            raise BadParamError

        buf.extend(param.encode('utf-8'))

    # Truncate if desired; leave 2 bytes over for \r\n:
    truncated = False
    if truncate_len is not None and (truncate_len - 2) < (len(buf) - len_tags):
        truncated = True
        new_buf_len = len_tags + (truncate_len - 2)
        buf = buf[:find_utf8_truncation_point(buf, new_buf_len)]

    buf.extend(b'\r\n')

    to_validate = buf[:-2]
    if any(c in to_validate for c in (b'\x00', b'\r', b'\n')):
        raise BadCharactersError

    raw = bytes(buf)

    return RenderedLine(
        raw=raw,
        truncated=truncated,
    )


def render_line(msg: Message) -> bytes:
    return render_line_(msg).raw


def render_line_strict(
        msg: Message,
        from_client: bool,
        truncate_len: int | None,
) -> RenderedLine:
    tag_limit: int | None = None
    client_only_tag_data_limit: int | None = None
    server_added_tag_data_limit: int | None = None
    if from_client:
        # enforce client max tags:
        # <client_max>   (4096)  :: '@' <tag_data 4094> ' '
        tag_limit = MAX_LEN_TAGS_FROM_CLIENT
    else:
        # on the server side, enforce separate client-only and server-added tag budgets:
        # "Servers MUST NOT add tag data exceeding 4094 bytes to messages."
        # <combined_max> (8191)  :: '@' <tag_data 4094> ';' <tag_data 4094> ' '
        client_only_tag_data_limit = MAX_LEN_CLIENT_TAG_DATA
        server_added_tag_data_limit = MAX_LEN_SERVER_TAG_DATA

    return render_line_(
        msg,
        tag_limit=tag_limit,
        client_only_tag_data_limit=client_only_tag_data_limit,
        server_added_tag_data_limit=server_added_tag_data_limit,
        truncate_len=truncate_len,
    )



import contextlib
import keyword
import sys
import types
import typing as ta

from .decorator import dataclass


##


def make_dataclass(  # noqa
        cls_name,
        fields,
        *,
        bases=(),
        namespace=None,

        init=True,
        repr=True,  # noqa
        eq=True,
        order=False,
        unsafe_hash=False,
        frozen=False,

        match_args=True,
        kw_only=False,
        slots=False,
        weakref_slot=False,

        module=None,

        #

        metadata: ta.Sequence[ta.Any] | None = None,

        reorder: bool | None = None,
        cache_hash: bool | None = None,
        generic_init: bool | None = None,
        override: bool | None = None,
        allow_dynamic_dunder_attrs: bool | None = None,

        repr_id: bool | None = None,
        terse_repr: bool | None = None,

        allow_redundant_decorator: bool | None = None,
):
    if namespace is None:
        namespace = {}

    seen = set()
    annotations = {}
    defaults = {}
    for item in fields:
        if isinstance(item, str):
            name = item
            tp = 'typing.Any'
        elif len(item) == 2:
            name, tp, = item
        elif len(item) == 3:
            name, tp, spec = item
            defaults[name] = spec
        else:
            raise TypeError(f'Invalid field: {item!r}')
        if not isinstance(name, str) or not name.isidentifier():
            raise TypeError(f'Field names must be valid identifiers: {name!r}')
        if keyword.iskeyword(name):
            raise TypeError(f'Field names must not be keywords: {name!r}')
        if name in seen:
            raise TypeError(f'Field name duplicated: {name!r}')

        seen.add(name)
        annotations[name] = tp

    def exec_body_callback(ns):
        ns.update(namespace)
        ns.update(defaults)
        ns['__annotations__'] = annotations

    cls = types.new_class(cls_name, bases, {}, exec_body_callback)

    if module is None:
        try:
            module = sys._getframemodulename(1) or '__main__'  # type: ignore  # noqa
        except AttributeError:
            with contextlib.suppress(AttributeError, ValueError):
                module = sys._getframe(1).f_globals.get('__name__', '__main__')  # noqa
    if module is not None:
        cls.__module__ = module

    return dataclass(
        cls,
        init=init,
        repr=repr,
        eq=eq,
        order=order,
        unsafe_hash=unsafe_hash,
        frozen=frozen,
        match_args=match_args,
        kw_only=kw_only,
        slots=slots,
        weakref_slot=weakref_slot,

        #

        metadata=metadata,

        reorder=reorder,
        cache_hash=cache_hash,
        generic_init=generic_init,
        override=override,
        allow_dynamic_dunder_attrs=allow_dynamic_dunder_attrs,

        repr_id=repr_id,
        terse_repr=terse_repr,

        allow_redundant_decorator=allow_redundant_decorator,
    )



import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh
from ... import reflect as rfl
from ...funcs import match as mfs
from .types import NotSpecified


##


_NOT_SPECIFIED_RTY = rfl.type_(type[NotSpecified])


@dc.dataclass(frozen=True)
class NotSpecifiedUnionMarshaler(msh.Marshaler):
    m: msh.Marshaler

    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        if o is NotSpecified:
            raise TypeError(o)
        return self.m.marshal(ctx, o)


class NotSpecifiedUnionMarshalerFactory(msh.MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: (
        isinstance(rty, rfl.Union) and
        not rty.is_optional and
        _NOT_SPECIFIED_RTY in rty.args
    ))
    def _build(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        args = set(check.isinstance(rty, rfl.Union).args) - {_NOT_SPECIFIED_RTY}
        nty = rfl.type_(ta.Union[*args])
        m = ctx.make(nty)
        return NotSpecifiedUnionMarshaler(m)


class NotSpecifiedUnionUnmarshalerFactory(msh.UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: (
        isinstance(rty, rfl.Union) and
        not rty.is_optional and
        _NOT_SPECIFIED_RTY in rty.args
    ))
    def _build(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        args = set(check.isinstance(rty, rfl.Union).args) - {_NOT_SPECIFIED_RTY}
        nty = rfl.type_(ta.Union[*args])
        return ctx.make(nty)


@lang.static_init
def _install_standard_marshalling() -> None:
    msh.install_standard_factories(
        msh.ForbiddenTypeMarshalerFactory({_NOT_SPECIFIED_RTY}),
        msh.ForbiddenTypeUnmarshalerFactory({_NOT_SPECIFIED_RTY}),

        NotSpecifiedUnionMarshalerFactory(),
        NotSpecifiedUnionUnmarshalerFactory(),
    )



from . import (  # noqa
    abc,
    copy,
    doc,
    eq,
    fields,
    frozen,
    hash,  # noqa
    init,
    matchargs,
    order,
    override,
    params,
    replace,
    repr,  # noqa
    slots,
)



# type: ignore
# ruff: noqa
# flake8: noqa
# @omlish-generated
# Generated from Json.g4 by ANTLR 4.13.2
from ....text.antlr._runtime._all import *
if "." in __name__:
    from .JsonParser import JsonParser
else:
    from JsonParser import JsonParser

# This class defines a complete listener for a parse tree produced by JsonParser.
class JsonListener(ParseTreeListener):

    # Enter a parse tree produced by JsonParser#json.
    def enterJson(self, ctx:JsonParser.JsonContext):
        pass

    # Exit a parse tree produced by JsonParser#json.
    def exitJson(self, ctx:JsonParser.JsonContext):
        pass


    # Enter a parse tree produced by JsonParser#obj.
    def enterObj(self, ctx:JsonParser.ObjContext):
        pass

    # Exit a parse tree produced by JsonParser#obj.
    def exitObj(self, ctx:JsonParser.ObjContext):
        pass


    # Enter a parse tree produced by JsonParser#pair.
    def enterPair(self, ctx:JsonParser.PairContext):
        pass

    # Exit a parse tree produced by JsonParser#pair.
    def exitPair(self, ctx:JsonParser.PairContext):
        pass


    # Enter a parse tree produced by JsonParser#arr.
    def enterArr(self, ctx:JsonParser.ArrContext):
        pass

    # Exit a parse tree produced by JsonParser#arr.
    def exitArr(self, ctx:JsonParser.ArrContext):
        pass


    # Enter a parse tree produced by JsonParser#value.
    def enterValue(self, ctx:JsonParser.ValueContext):
        pass

    # Exit a parse tree produced by JsonParser#value.
    def exitValue(self, ctx:JsonParser.ValueContext):
        pass



del JsonParser


import typing as ta

from ... import cached
from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base import MarshalContext
from ..base import Marshaler
from ..base import SimpleMarshalerFactory
from ..base import SimpleUnmarshalerFactory
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..values import Value
from .marshal import make_polymorphism_marshaler
from .metadata import Impls
from .metadata import TypeTagging
from .metadata import WrapperTypeTagging
from .unmarshal import make_polymorphism_unmarshaler


##


class MatchUnionMarshaler(Marshaler):
    mmf: mfs.MultiMatchFn[[UnmarshalContext, Value], ta.Any]

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        try:
            m = self.mmf.match(ctx, o)
        except mfs.AmbiguousMatchesError:
            raise ValueError(o)  # noqa
        return m.fn(ctx, o)


class MatchUnionUnmarshaler(Unmarshaler):
    mmf: mfs.MultiMatchFn[[UnmarshalContext, Value], ta.Any]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        try:
            m = self.mmf.match(ctx, v)
        except mfs.AmbiguousMatchesError:
            raise ValueError(v)  # noqa
        return m.fn(ctx, v)


##


PRIMITIVE_UNION_TYPES: tuple[type, ...] = (
    float,
    int,
    str,
)


#


@dc.dataclass(frozen=True)
class PrimitiveUnionMarshaler(Marshaler):
    tys: ta.Sequence[type]

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        if type(o) not in self.tys:
            raise TypeError(o)
        return o


@dc.dataclass(frozen=True)
class PrimitiveUnionMarshalerFactory(SimpleMarshalerFactory):
    tys: ta.Sequence[type] = PRIMITIVE_UNION_TYPES

    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Union) and all(a in self.tys for a in rty.args)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        args = check.isinstance(rty, rfl.Union).args
        return PrimitiveUnionMarshaler([t for t in self.tys if t in args])


#


@dc.dataclass(frozen=True)
class PrimitiveUnionUnmarshaler(Unmarshaler):
    tys: ta.Sequence[type]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        if type(v) not in self.tys:
            raise TypeError(v)
        return v


@dc.dataclass(frozen=True)
class PrimitiveUnionUnmarshalerFactory(SimpleUnmarshalerFactory):
    tys: ta.Sequence[type] = PRIMITIVE_UNION_TYPES

    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Union) and all(a in self.tys for a in rty.args)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        args = check.isinstance(rty, rfl.Union).args
        return PrimitiveUnionUnmarshaler([t for t in self.tys if t in args])


#


PRIMITIVE_UNION_MARSHALER_FACTORY = PrimitiveUnionMarshalerFactory()
PRIMITIVE_UNION_UNMARSHALER_FACTORY = PrimitiveUnionUnmarshalerFactory()


##


@dc.dataclass(frozen=True)
class _BasePolymorphismUnionFactory(lang.Abstract):
    impls: Impls
    tt: TypeTagging = WrapperTypeTagging()
    allow_partial: bool = dc.field(default=False, kw_only=True)

    @cached.property
    @dc.init
    def rtys(self) -> frozenset[rfl.Type]:
        return frozenset(i.ty for i in self.impls)

    def guard(self, ctx: MarshalContext | UnmarshalContext, rty: rfl.Type) -> bool:
        if not isinstance(rty, rfl.Union):
            return False
        if self.allow_partial:
            return not (rty.args - self.rtys)
        else:
            return rty.args == self.rtys

    def get_impls(self, rty: rfl.Type) -> Impls:
        uty = check.isinstance(rty, rfl.Union)
        return Impls([self.impls.by_ty[check.isinstance(a, type)] for a in uty.args])


@dc.dataclass(frozen=True)
class PolymorphismUnionMarshalerFactory(_BasePolymorphismUnionFactory, SimpleMarshalerFactory):
    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return make_polymorphism_marshaler(self.get_impls(rty), self.tt, ctx)


@dc.dataclass(frozen=True)
class PolymorphismUnionUnmarshalerFactory(_BasePolymorphismUnionFactory, SimpleUnmarshalerFactory):
    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return make_polymorphism_unmarshaler(self.get_impls(rty), self.tt, ctx)



import logging


##


NOISY_LOGGERS: set[str] = {
    'boto3.resources.action',
    'datadog.dogstatsd',
    'elasticsearch',
    'httpx',
    'kazoo.client',
    'requests.packages.urllib3.connectionpool',
}


def silence_noisy_loggers() -> None:
    for noisy_logger in NOISY_LOGGERS:
        logging.getLogger(noisy_logger).setLevel(logging.WARNING)



import concurrent.futures as cf
import contextlib
import typing as ta


T = ta.TypeVar('T')
P = ta.ParamSpec('P')


##


class ImmediateExecutor(cf.Executor):
    def __init__(self, *, immediate_exceptions: bool = False) -> None:
        super().__init__()
        self._immediate_exceptions = immediate_exceptions

    def submit(
            self,
            fn: ta.Callable[P, T],
            /,
            *args: P.args,
            **kwargs: P.kwargs,
    ) -> cf.Future[T]:
        future: ta.Any = cf.Future()
        try:
            result = fn(*args, **kwargs)
            future.set_result(result)
        except Exception as e:
            if self._immediate_exceptions:
                raise
            future.set_exception(e)
        return future


@contextlib.contextmanager
def new_executor(
        max_workers: int | None = None,
        cls: type[cf.Executor] = cf.ThreadPoolExecutor,
        *,
        immediate_exceptions: bool = False,
        **kwargs: ta.Any,
) -> ta.Generator[cf.Executor]:
    if max_workers == 0:
        yield ImmediateExecutor(
            immediate_exceptions=immediate_exceptions,
        )

    else:
        with cls(  # type: ignore
                max_workers,
                **kwargs,
        ) as exe:
            yield exe



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import asyncio
import typing as ta


##


ASYNCIO_DEFAULT_BUFFER_LIMIT = 2 ** 16


async def asyncio_open_stream_reader(
        f: ta.IO,
        loop: ta.Any = None,
        *,
        limit: int = ASYNCIO_DEFAULT_BUFFER_LIMIT,
) -> asyncio.StreamReader:
    if loop is None:
        loop = asyncio.get_running_loop()

    reader = asyncio.StreamReader(limit=limit, loop=loop)
    await loop.connect_read_pipe(
        lambda: asyncio.StreamReaderProtocol(reader, loop=loop),
        f,
    )

    return reader


async def asyncio_open_stream_writer(
        f: ta.IO,
        loop: ta.Any = None,
) -> asyncio.StreamWriter:
    if loop is None:
        loop = asyncio.get_running_loop()

    writer_transport, writer_protocol = await loop.connect_write_pipe(
        lambda: asyncio.streams.FlowControlMixin(loop=loop),
        f,
    )

    return asyncio.streams.StreamWriter(
        writer_transport,
        writer_protocol,
        None,
        loop,
    )



import typing as ta

from .... import check
from ...internals import STD_PARAMS_ATTR
from ...internals import StdParams
from ...specs import ClassSpec


##


STD_PARAM_DEFAULTS: ta.Mapping[str, ta.Any] = dict(
    init=True,
    repr=True,  # noqa
    eq=True,
    order=False,
    unsafe_hash=False,
    frozen=False,

    match_args=True,
    kw_only=False,
    slots=False,
    weakref_slot=False,
)

STD_PARAM_ATTRS: tuple[str, ...] = tuple(STD_PARAM_DEFAULTS)


##


class SpecDataclassParams(StdParams):
    __slots__ = (*StdParams.__slots__, '_spec')

    def __init__(
            self,
            init,
            repr,  # noqa
            eq,
            order,
            unsafe_hash,
            frozen,
            match_args,
            kw_only,
            slots,
            weakref_slot,
            *args: ta.Any,
            spec: ClassSpec,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(
            init,
            repr,
            eq,
            order,
            unsafe_hash,
            frozen,
            match_args,
            kw_only,
            slots,
            weakref_slot,
            *args,
            **kwargs,
        )

        self._spec = spec

    def __repr__(self) -> str:
        r = super().__repr__()
        return f'{self.__class__.__name__}{r[r.index("("):]}'


def build_spec_std_params(cs: ClassSpec) -> 'SpecDataclassParams':
    return SpecDataclassParams(
        init=cs.init,
        repr=cs.repr,
        eq=cs.eq,
        order=cs.order,
        unsafe_hash=cs.unsafe_hash,
        frozen=cs.frozen,

        match_args=cs.match_args,
        kw_only=cs.kw_only,
        slots=cs.slots,
        weakref_slot=cs.weakref_slot,

        spec=cs,
    )


def get_class_spec(cls: type) -> ClassSpec | None:
    check.isinstance(cls, type)

    sp = getattr(cls, STD_PARAMS_ATTR)
    if not isinstance(sp, SpecDataclassParams):
        return None

    return check.isinstance(sp._spec, ClassSpec)  # noqa



# ruff: noqa: UP007 UP012 UP045
# @omlish-lite
import ctypes as ct
import logging
import sys
import syslog
import threading
import typing as ta

from ..lite.cached import cached_nullary


##


class sd_iovec(ct.Structure):  # noqa
    pass


sd_iovec._fields_ = [
    ('iov_base', ct.c_void_p),  # Pointer to data.
    ('iov_len', ct.c_size_t),  # Length of data.
]


##


@cached_nullary
def sd_libsystemd() -> ta.Any:
    lib = ct.CDLL('libsystemd.so.0')

    lib.sd_journal_sendv.restype = ct.c_int
    lib.sd_journal_sendv.argtypes = [ct.POINTER(sd_iovec), ct.c_int]

    return lib


@cached_nullary
def sd_try_libsystemd() -> ta.Optional[ta.Any]:
    try:
        return sd_libsystemd()
    except OSError:  # noqa
        return None


##


def sd_journald_send(**fields: str) -> int:
    lib = sd_libsystemd()

    msgs = [
        f'{k.upper()}={v}\0'.encode('utf-8')
        for k, v in fields.items()
    ]

    vec = (sd_iovec * len(msgs))()
    cl = (ct.c_char_p * len(msgs))()  # noqa
    for i in range(len(msgs)):
        vec[i].iov_base = ct.cast(ct.c_char_p(msgs[i]), ct.c_void_p)
        vec[i].iov_len = len(msgs[i]) - 1

    return lib.sd_journal_sendv(vec, len(msgs))


##


SD_LOG_LEVEL_MAP: ta.Mapping[int, int] = {
    logging.FATAL: syslog.LOG_EMERG,  # system is unusable
    # LOG_ALERT ?  # action must be taken immediately
    logging.CRITICAL: syslog.LOG_CRIT,
    logging.ERROR: syslog.LOG_ERR,
    logging.WARNING: syslog.LOG_WARNING,
    # LOG_NOTICE ?  # normal but significant condition
    logging.INFO: syslog.LOG_INFO,
    logging.DEBUG: syslog.LOG_DEBUG,
}


class JournaldLogHandler(logging.Handler):
    """
    TODO:
     - fallback handler for when this barfs
    """

    def __init__(
            self,
            *,
            use_formatter_output: bool = False,
    ) -> None:
        super().__init__()

        sd_libsystemd()

        self._use_formatter_output = use_formatter_output

    #

    EXTRA_RECORD_ATTRS_BY_JOURNALD_FIELD: ta.ClassVar[ta.Mapping[str, str]] = {
        'name': 'name',
        'module': 'module',
        'exception': 'exc_text',
        'thread_name': 'threadName',
        'task_name': 'taskName',
    }

    def make_fields(self, record: logging.LogRecord) -> ta.Mapping[str, str]:
        formatter_message = self.format(record)
        if self._use_formatter_output:
            message = formatter_message
        else:
            message = record.message

        fields: dict[str, str] = {
            'message': message,
            'priority': str(SD_LOG_LEVEL_MAP[record.levelno]),
            'tid': str(threading.get_ident()),
        }

        if (pathname := record.pathname) is not None:
            fields['code_file'] = pathname
        if (lineno := record.lineno) is not None:
            fields['code_lineno'] = str(lineno)
        if (func_name := record.funcName) is not None:
            fields['code_func'] = func_name

        for f, a in self.EXTRA_RECORD_ATTRS_BY_JOURNALD_FIELD.items():
            if (v := getattr(record, a, None)) is not None:
                fields[f] = str(v)

        return fields

    #

    def emit(self, record: logging.LogRecord) -> None:
        try:
            fields = self.make_fields(record)

            if rc := sd_journald_send(**fields):
                raise RuntimeError(f'{self.__class__.__name__}.emit failed: {rc=}')  # noqa

        except RecursionError:  # See issue 36272
            raise

        except Exception:  # noqa
            self.handleError(record)


def journald_log_handler_factory(
        *,
        no_tty_check: bool = False,
        no_fallback: bool = False,
) -> logging.Handler:
    if (
            sys.platform == 'linux' and
            (no_tty_check or not sys.stderr.isatty()) and
            (no_fallback or sd_try_libsystemd() is not None)
    ):
        return JournaldLogHandler()

    return logging.StreamHandler()



import typing as ta

from ..lite.maysyncs import make_maysync as _make_maysync
from ..lite.maysyncs import maysync as _maysync


T_co = ta.TypeVar('T_co', covariant=True)
P = ta.ParamSpec('P')


##


class MaysyncableP(ta.Protocol[P, T_co]):
    @property
    def s(self) -> ta.Callable[P, T_co]:
        ...

    @property
    def a(self) -> ta.Callable[P, ta.Awaitable[T_co]]:
        ...

    @property
    def m(self) -> ta.Callable[P, ta.Awaitable[T_co]]:
        ...


def make_maysync(
        s: ta.Callable[P, T_co],
        a: ta.Callable[P, ta.Awaitable[T_co]],
) -> MaysyncableP[P, T_co]:
    return ta.cast('MaysyncableP[P, T_co]', _make_maysync(
        s,
        a,
    ))


def maysync(m: ta.Callable[P, ta.Awaitable[T_co]]) -> MaysyncableP[P, T_co]:
    return ta.cast('MaysyncableP[P, T_co]', _maysync(
        m,
    ))



"""
TODO:
 - zipfile?
 - tarfile?
 - lzf?
 - blosc?
 - python-lzo?
 - zfp -> fp8?

Compression choice:
 - lzma if-available minimal-space
 - lz4 if-available read-heavy
 - zstd if-available
 - bz2 read-heavy (but no parallel decompress)
 - gz
"""



# ruff: noqa: UP006 UP007 UP045
"""
https://docs.python.org/3/library/unittest.html#command-line-interface
~ https://github.com/python/cpython/tree/f66c75f11d3aeeb614600251fd5d3fe1a34b5ff1/Lib/unittest

TODO:
 - giving it filenames doesn't work
"""
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001-2024 Python Software Foundation; All Rights
# Reserved" are retained in Python alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import argparse
import dataclasses as dc
import importlib.util
import os
import re
import sys
import types
import typing as ta

from .loading import UnittestTargetLoader
from .running import UnittestTestRunner


##


class UnittestRunCli:
    def __init__(self) -> None:
        super().__init__()

        self._parser = self._get_arg_parser()

    #

    DEFAULT_DISCOVERY_START = '.'
    DEFAULT_DISCOVERY_PATTERN = 'test*.py'

    def _get_arg_parser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(add_help=False)

        parser.add_argument(
            '-v',
            '--verbose',
            dest='verbosity',
            action='store_const',
            const=2,
            help='Verbose output',
        )

        parser.add_argument(
            '-q',
            '--quiet',
            dest='verbosity',
            action='store_const',
            const=0,
            help='Quiet output',
        )

        parser.add_argument(
            '--locals',
            dest='tb_locals',
            action='store_true',
            help='Show local variables in tracebacks',
        )

        parser.add_argument(
            '-f',
            '--failfast',
            dest='failfast',
            action='store_true',
            help='Stop on first fail or error',
        )

        parser.add_argument(
            '-c',
            '--catch',
            dest='catchbreak',
            action='store_true',
            help='Catch Ctrl-C and display results so far',
        )

        parser.add_argument(
            '-b',
            '--buffer',
            dest='buffer',
            action='store_true',
            help='Buffer stdout and stderr during tests',
        )

        def _convert_select_pattern(pattern: str) -> str:
            if '*' not in pattern:
                pattern = f'*{pattern}*'
            return pattern

        parser.add_argument(
            '-k',
            dest='test_name_patterns',
            action='append',
            type=_convert_select_pattern,
            help='Only run tests which match the given substring',
        )

        #

        parser.epilog = (
            'For test discovery all test modules must be importable from the top level directory of the project.'
        )

        parser.add_argument(
            '-s',
            '--start-directory',
            default=self.DEFAULT_DISCOVERY_START,
            dest='start',
            help="Directory to start discovery ('.' default)",
        )

        parser.add_argument(
            '-p',
            '--pattern',
            default=self.DEFAULT_DISCOVERY_PATTERN,
            dest='pattern',
            help="Pattern to match tests ('test*.py' default)",
        )

        parser.add_argument(
            '-t',
            '--top-level-directory',
            dest='top',
            help='Top level directory of project (defaults to start directory)',
        )

        parser.add_argument(
            'target',
            nargs='*',
            default=argparse.SUPPRESS,
            help=argparse.SUPPRESS,
        )

        return parser

    #

    @dc.dataclass(frozen=True)
    class ParsedArgs:
        args: ta.Optional[argparse.Namespace]
        test_names: ta.Optional[ta.Sequence[str]] = None
        module: ta.Union[str, types.ModuleType, None] = None

    def parse_args(self, argv: ta.Sequence[str]) -> ParsedArgs:
        args = self._parser.parse_args(argv)
        return self.ParsedArgs(args)

    #

    @staticmethod
    def _get_attr_dict(obj: ta.Optional[ta.Any], *attrs: str) -> ta.Dict[str, ta.Any]:
        if obj is None:
            return {}

        return {
            a: v
            for a in attrs
            if (v := getattr(obj, a, None)) is not None
        }

    #

    IMPORT_PATH_PAT = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*(?:\.[a-zA-Z_][a-zA-Z0-9_]*)*')

    def _build_target(self, name: str, args: ParsedArgs) -> UnittestTargetLoader.Target:
        is_discovery = False
        if os.path.isdir(name):
            is_discovery = True
        elif self.IMPORT_PATH_PAT.fullmatch(name):
            spec = importlib.util.find_spec(name)
            if spec is not None and spec.submodule_search_locations is not None:  # is a package, not a module
                is_discovery = True

        if not is_discovery:
            return UnittestTargetLoader.NamesTarget([name])

        else:
            return UnittestTargetLoader.DiscoveryTarget(
                start=name,
                **self._get_attr_dict(
                    args.args,
                    'pattern',
                    'top',
                ),
            )

    #

    NO_TESTS_EXITCODE = 5

    def run(
            self,
            args: ParsedArgs,
            *,
            exit: bool = False,  # noqa
    ) -> None:
        loader = UnittestTargetLoader(**self._get_attr_dict(
            args.args,
            'test_name_patterns',
        ))

        tests = [
            loader.load(self._build_target(target_arg, args))
            for target_arg in (args.args.target if args.args is not None else None) or []  # noqa
        ]

        runner = UnittestTestRunner(UnittestTestRunner.Args(**self._get_attr_dict(
            args.args,
            'verbosity',
            'failfast',
            'catchbreak',
            'buffer',
            'warnings',
            'tb_locals',
        )))

        result = runner.run_many(tests)

        runner.print(result)

        if exit:
            if not result.num_tests_run and not result.skipped:
                sys.exit(self.NO_TESTS_EXITCODE)
            elif result.was_successful:
                sys.exit(0)
            else:
                sys.exit(1)


##


def _main() -> None:
    cli = UnittestRunCli()
    args = cli.parse_args(sys.argv[1:])
    cli.run(args, exit=True)


if __name__ == '__main__':
    _main()



from .base import (  # noqa
    Marshaler,
    Unmarshaler,

    MarshalerMaker,
    UnmarshalerMaker,

    MarshalerFactory,
    UnmarshalerFactory,

    MarshalerFactory_,
    UnmarshalerFactory_,

    SimpleMarshalerFactory,
    SimpleUnmarshalerFactory,

    MarshalerFactoryMatchClass,
    UnmarshalerFactoryMatchClass,

    MultiMarshalerFactory,
    MultiUnmarshalerFactory,

    TypeMapMarshalerFactory,
    TypeMapUnmarshalerFactory,

    TypeCacheMarshalerFactory,
    TypeCacheUnmarshalerFactory,

    FuncMarshaler,
    FuncUnmarshaler,

    BaseContext,
    MarshalContext,
    UnmarshalContext,

    RecursiveMarshalerFactory,
    RecursiveUnmarshalerFactory,

    Override,
    ReflectOverride,
)

from .composite.iterables import (  # noqa
    IterableMarshaler,
    IterableUnmarshaler,
)

from .composite.wrapped import (  # noqa
    WrappedMarshaler,
    WrappedUnmarshaler,
)

from .errors import (  # noqa
    ForbiddenTypeError,
    MarshalError,
    UnhandledTypeError,
)

from .global_ import (  # noqa
    GLOBAL_REGISTRY,

    global_marshaler_factory,
    marshal,

    global_unmarshaler_factory,
    unmarshal,

    register_global,
)

from .naming import (  # noqa
    Naming,
    translate_name,
)

from .objects.dataclasses import (  # noqa
    AbstractDataclassFactory,
    DataclassMarshalerFactory,
    DataclassUnmarshalerFactory,
    get_dataclass_field_infos,
    get_dataclass_metadata,
)

from .objects.helpers import (  # noqa
    update_fields_metadata,
    update_object_metadata,
    with_field_metadata,
)

from .objects.marshal import (  # noqa
    ObjectMarshaler,
    SimpleObjectMarshalerFactory,
)

from .objects.metadata import (  # noqa
    FieldInfo,
    FieldInfos,
    FieldMetadata,
    FieldOptions,
    ObjectMetadata,
    ObjectSpecials,
)

from .objects.unmarshal import (  # noqa
    ObjectUnmarshaler,
    SimpleObjectUnmarshalerFactory,
)

from .polymorphism.marshal import (  # noqa
    PolymorphismMarshaler,
    PolymorphismMarshalerFactory,
    make_polymorphism_marshaler,
)

from .polymorphism.metadata import (  # noqa
    FieldTypeTagging,
    Impl,
    Impls,
    Polymorphism,
    TypeTagging,
    WrapperTypeTagging,
    polymorphism_from_impls,
    polymorphism_from_subclasses,
)

from .polymorphism.standard import (  # noqa
    standard_polymorphism_factories,
)

from .polymorphism.unions import (  # noqa
    PRIMITIVE_UNION_TYPES,
    PolymorphismUnionMarshalerFactory,
    PolymorphismUnionUnmarshalerFactory,
    PrimitiveUnionMarshaler,
    PrimitiveUnionMarshalerFactory,
    PrimitiveUnionUnmarshaler,
    PrimitiveUnionUnmarshalerFactory,
)

from .polymorphism.unmarshal import (  # noqa
    PolymorphismUnmarshaler,
    PolymorphismUnmarshalerFactory,
    make_polymorphism_unmarshaler,
)

from .singular.base64 import (  # noqa
    BASE64_MARSHALER_FACTORY,
    BASE64_UNMARSHALER_FACTORY,
)

from .singular.primitives import (  # noqa
    PRIMITIVE_TYPES,
)

from .registries import (  # noqa
    Registry,
)

from .standard import (  # noqa
    STANDARD_MARSHALER_FACTORIES,
    new_standard_marshaler_factory,

    STANDARD_UNMARSHALER_FACTORIES,
    new_standard_unmarshaler_factory,

    install_standard_factories,
)

from .trivial.forbidden import (  # noqa
    ForbiddenTypeMarshalerFactory,
    ForbiddenTypeUnmarshalerFactory,
)

from .trivial.nop import (  # noqa
    NOP_MARSHALER_UNMARSHALER,
    NopMarshalerUnmarshaler,
)

from .values import (  # noqa
    Value,
)


##


from ..lang.imports import _trigger_conditional_imports  # noqa

_trigger_conditional_imports(__package__)



import typing as ta

from ... import check
from ... import lang
from ..buffers import ReadableListBuffer
from .consts import DEFAULT_BUFFER_SIZE
from .direct import BytesDirectCoro
from .direct import StrDirectCoro


T = ta.TypeVar('T')

O = ta.TypeVar('O')
I = ta.TypeVar('I')
R = ta.TypeVar('R')

OF = ta.TypeVar('OF')
OT = ta.TypeVar('OT')


# Stepped coros accept a non-None input, then in response yield zero or more non-None outputs, until yielding None to
# signal they need more input again.
SteppedCoro: ta.TypeAlias = ta.Generator[O | None, I | None, R]

# Conventionally, these are sent and themselves yield an empty value to signify termination.
BytesSteppedCoro: ta.TypeAlias = SteppedCoro[bytes, bytes, R]
StrSteppedCoro: ta.TypeAlias = SteppedCoro[str, str, R]

BytesToStrSteppedCoro: ta.TypeAlias = SteppedCoro[str, bytes, R]
StrToBytesSteppedCoro: ta.TypeAlias = SteppedCoro[bytes, str, R]


# Stepped reader generators emit either an int or None to request input, or emit some other kind of output.
SteppedReaderCoro: ta.TypeAlias = ta.Generator[int | None | O, I | None, R]

BytesSteppedReaderCoro: ta.TypeAlias = SteppedReaderCoro[bytes, bytes, R]
StrSteppedReaderCoro: ta.TypeAlias = SteppedReaderCoro[str, str, R]


##


@lang.autostart
def flatmap_stepped_coro(
        fn: ta.Callable[[list[OF]], OT],
        g: SteppedCoro[OF, I, R],
        *,
        terminate: ta.Callable[[OF], bool] | None = None,
) -> ta.Generator[OT, I, lang.Maybe[R]]:
    """
    Given a stepped coro and a function taking a list, returns a direct (1:1) coro which accepts input, builds a list of
    yielded coro output, calls the given function with that list, and yields the result.

    An optional terminate function may be provided which will cause this function to return early if it returns true for
    an encountered yielded value. The encountered value causing termination will be included in the list sent to the
    given fn.

    Returns a Maybe of either the given coro's return value or empty if the terminator was encountered.
    """

    l: list[OF]
    i: I | None = yield  # type: ignore
    while True:
        l = []

        while True:
            try:
                o = g.send(i)
            except StopIteration as e:
                if l:
                    yield fn(l)
                return lang.just(e.value)

            i = None

            if o is None:
                break

            l.append(o)

            if terminate is not None and terminate(o):
                yield fn(l)
                return lang.empty()

        i = yield fn(l)


##


def _join_bytes(l: ta.Sequence[bytes]) -> bytes:
    if not l:
        return b''
    elif len(l) == 1:
        return l[0]
    else:
        return b''.join(l)


def _join_str(l: ta.Sequence[str]) -> str:
    if not l:
        return ''
    elif len(l) == 1:
        return l[0]
    else:
        return ''.join(l)


def _is_empty(o: T) -> bool:
    return len(o) < 1  # type: ignore


def joined_bytes_stepped_coro(g: BytesSteppedCoro[R]) -> BytesDirectCoro[R]:
    return flatmap_stepped_coro(_join_bytes, g, terminate=_is_empty)


def joined_str_stepped_coro(g: StrSteppedCoro[R]) -> StrDirectCoro[R]:
    return flatmap_stepped_coro(_join_str, g, terminate=_is_empty)


##


def read_into_bytes_stepped_coro(
        g: BytesSteppedCoro,
        f: ta.IO,
        *,
        read_size: int = DEFAULT_BUFFER_SIZE,
) -> ta.Iterator[bytes]:
    yield from lang.genmap(  # type: ignore[misc]
        joined_bytes_stepped_coro(g),
        lang.readiter(f, read_size),
    )


def read_into_str_stepped_coro(
        g: StrSteppedCoro,
        f: ta.TextIO,
        *,
        read_size: int = DEFAULT_BUFFER_SIZE,
) -> ta.Iterator[str]:
    yield from lang.genmap(
        joined_str_stepped_coro(g),
        lang.readiter(f, read_size),
    )


##


@lang.autostart
def buffer_bytes_stepped_reader_coro(g: BytesSteppedReaderCoro) -> BytesSteppedCoro:
    i: bytes | None
    o = g.send(None)
    rlb = ReadableListBuffer()
    eof = False

    while True:
        if eof:
            raise EOFError

        if not len(rlb):
            if (more := check.isinstance((yield None), bytes)):
                rlb.feed(more)
            else:
                eof = True

        if o is None:
            i = check.not_none(rlb.read())

        elif isinstance(o, int):
            while len(rlb) < o:
                more = check.isinstance((yield None), bytes)
                if not more:
                    raise EOFError
                # FIXME: lol - share guts with readers
                rlb.feed(more)

            i = check.not_none(rlb.read(o))

        else:
            raise TypeError(o)

        while True:
            o = g.send(i)
            i = None
            if isinstance(o, bytes):
                check.none((yield o))
                if not o:
                    return
            else:
                break


##


@lang.autostart
def iterable_bytes_stepped_coro(g: BytesSteppedCoro) -> ta.Generator[ta.Iterator[bytes], bytes]:
    i: bytes | None = check.isinstance((yield None), bytes)  # type: ignore[misc]
    eof = False

    def f() -> ta.Iterator[bytes]:
        nonlocal i
        while True:
            o = g.send(i)
            i = None

            if isinstance(o, bytes):
                yield o
                if not o:
                    nonlocal eof
                    eof = True
                    return
            elif o is None:
                return
            else:
                raise TypeError(o)

    while True:
        if eof:
            return

        i = (yield f())  # noqa



import dataclasses as dc
import typing as ta

from ..base import MarshalContext
from ..base import Marshaler
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..values import Value


##


@dc.dataclass(frozen=True)
class WrappedMarshaler(Marshaler):
    wrapper: ta.Callable[[MarshalContext, ta.Any], ta.Any]
    m: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return self.m.marshal(ctx, self.wrapper(ctx, o))


@dc.dataclass(frozen=True)
class WrappedUnmarshaler(Unmarshaler):
    unwrapper: ta.Callable[[UnmarshalContext, ta.Any], ta.Any]
    u: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return self.unwrapper(ctx, self.u.unmarshal(ctx, v))



# ruff: noqa: N802 N803
import typing as ta

from ... import check
from ...text.antlr import runtime as antlr4
from ...text.antlr.errors import SilentRaisingErrorListener
from . import nodes as no
from ._antlr.Protobuf3Lexer import Protobuf3Lexer  # type: ignore
from ._antlr.Protobuf3Parser import Protobuf3Parser  # type: ignore
from ._antlr.Protobuf3Visitor import Protobuf3Visitor  # type: ignore


##


class _ParseVisitor(Protobuf3Visitor):
    def visit(self, ctx: antlr4.ParserRuleContext):
        check.isinstance(ctx, antlr4.ParserRuleContext)
        node = ctx.accept(self)
        return node

    def aggregateResult(self, aggregate, nextResult):
        if aggregate is not None:
            check.none(nextResult)
            return aggregate
        else:
            check.none(aggregate)
            return nextResult

    #

    def visitField(self, ctx: Protobuf3Parser.FieldContext):
        name = ctx.fieldName().getText()
        type_: no.Type = check.isinstance(self.visitTypeRule(ctx.typeRule()), (no.SimpleType, no.TypeRef))
        num = int(ctx.fieldNumber().getText())
        repeated = ctx.REPEATED() is not None
        return no.Field(
            name=name,
            type=type_,
            num=num,
            repeated=repeated,
        )

    def visitMessage(self, ctx: Protobuf3Parser.MessageContext):
        name = ctx.messageName().getText()
        fields: list[no.Field] = []
        for mbc_ctx in ctx.messageBody().messageBodyContent():
            mbc = self.visitMessageBodyContent(mbc_ctx)
            if isinstance(mbc, no.Field):
                fields.append(mbc)
        return no.Message(
            name=name,
            fields=fields,
        )

    def visitMessageOrEnumType(self, ctx: Protobuf3Parser.MessageOrEnumTypeContext):
        return no.TypeRef(ctx.getText())

    def visitProto(self, ctx: Protobuf3Parser.ProtoContext):
        messages: list[no.Message] = []
        for ex_ctx in ctx.syntaxExtra():
            ex = self.visitSyntaxExtra(ex_ctx)
            if isinstance(ex, no.Message):
                messages.append(ex)
        return no.ProtoFile(
            messages=messages,
        )

    def visitSimpleType(self, ctx: Protobuf3Parser.SimpleTypeContext):
        return no.SimpleType(ctx.getText())


##


def create_parser(buf: str) -> Protobuf3Parser:
    lexer = Protobuf3Lexer(antlr4.InputStream(buf))
    lexer.removeErrorListeners()
    lexer.addErrorListener(SilentRaisingErrorListener())

    stream = antlr4.CommonTokenStream(lexer)
    stream.fill()

    parser = Protobuf3Parser(stream)
    parser.removeErrorListeners()
    parser.addErrorListener(SilentRaisingErrorListener())

    return parser


##


def parse_proto(buf: str, **kwargs: ta.Any) -> ta.Any:
    parser = create_parser(buf, **kwargs)
    node = _ParseVisitor().visit(parser.proto())
    return node



# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
# ~> https://github.com/python/cpython/blob/f19c50a4817ffebb26132182ed8cec6a72342cc0/Lib/_compression.py
import typing as ta

from ... import check
from ..coro import BytesSteppedCoro
from ..coro import BytesSteppedReaderCoro
from .abc import CompressorObject
from .abc import NeedsInputDecompressorObject
from .abc import UnconsumedTailDecompressorObject


##


class CompressorObjectIncrementalAdapter:
    def __init__(
            self,
            factory: ta.Callable[..., CompressorObject],
    ) -> None:
        super().__init__()

        self._factory = factory

    def __call__(self) -> BytesSteppedCoro:
        compressor = self._factory()

        while True:
            data = check.isinstance((yield None), bytes)
            if not data:
                break

            compressed = compressor.compress(data)
            if compressed:
                check.none((yield compressed))

        if (fl := compressor.flush()):
            check.none((yield fl))

        check.none((yield b''))


##


class DecompressorObjectIncrementalAdapter:
    def __init__(
            self,
            factory: ta.Callable[..., NeedsInputDecompressorObject | UnconsumedTailDecompressorObject],
            *,
            trailing_error: type[BaseException] | tuple[type[BaseException], ...] = (),
    ) -> None:
        super().__init__()

        self._factory = factory
        self._trailing_error = trailing_error

    def __call__(self) -> BytesSteppedReaderCoro:
        pos = 0

        data = None  # Default if EOF is encountered

        decompressor = self._factory()

        while True:
            # Depending on the input data, our call to the decompressor may not return any data. In this case, try again
            # after reading another block.
            while True:
                if decompressor.eof:
                    rawblock = decompressor.unused_data
                    if not rawblock:
                        rawblock = check.isinstance((yield None), bytes)
                    if not rawblock:
                        break

                    # Continue to next stream.
                    decompressor = self._factory()

                    try:
                        data = decompressor.decompress(rawblock)
                    except self._trailing_error:
                        # Trailing data isn't a valid compressed stream; ignore it.
                        break

                else:
                    if hasattr(decompressor, 'needs_input'):
                        if decompressor.needs_input:
                            rawblock = check.isinstance((yield None), bytes)
                            if not rawblock:
                                raise EOFError('Compressed file ended before the end-of-stream marker was reached')
                        else:
                            rawblock = b''

                    elif hasattr(decompressor, 'unconsumed_tail'):
                        if not (rawblock := decompressor.unconsumed_tail):
                            rawblock = check.isinstance((yield None), bytes)
                            if not rawblock:
                                raise EOFError('Compressed file ended before the end-of-stream marker was reached')

                    else:
                        raise TypeError(decompressor)

                    data = decompressor.decompress(rawblock)

                if data:
                    break

            if not data:
                check.none((yield b''))
                return

            pos += len(data)
            check.none((yield data))
            data = None



# ruff: noqa: UP045
# @omlish-lite
import abc
import dataclasses as dc
import functools
import typing as ta

from .args import Args
from .check import check


T = ta.TypeVar('T')

_MaysyncGen = ta.Generator['_MaysyncOp', ta.Any, T]  # ta.TypeAlias


##


@dc.dataclass(frozen=True, eq=False)
class Maysyncable(abc.ABC, ta.Generic[T]):
    @abc.abstractmethod
    def s(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def a(self, *args: ta.Any, **kwargs: ta.Any) -> ta.Awaitable[T]:
        raise NotImplementedError

    @ta.final
    def m(self, *args: ta.Any, **kwargs: ta.Any) -> ta.Awaitable[T]:
        return _MaysyncFuture(_MaysyncOp(self, Args(*args, **kwargs)))


##


@dc.dataclass(frozen=True, eq=False)
class _FnMaysyncable(Maysyncable[T]):
    s: ta.Callable[..., T]
    a: ta.Callable[..., ta.Awaitable[T]]

    def __post_init__(self) -> None:
        check.not_none(self.s)
        check.not_none(self.a)

    def __init_subclass__(cls, **kwargs):
        raise TypeError


_FnMaysyncable.__abstractmethods__ = frozenset()


def make_maysync(
        s: ta.Callable[..., T],
        a: ta.Callable[..., ta.Awaitable[T]],
) -> Maysyncable[T]:
    return _FnMaysyncable(
        s,
        a,
    )


##


@dc.dataclass(frozen=True, eq=False)
class _MgMaysyncable(Maysyncable[T]):
    mg: ta.Callable[..., _MaysyncGen[T]]

    def __init_subclass__(cls, **kwargs):
        raise TypeError

    def s(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        g = self.mg(*args, **kwargs)

        i: ta.Any = None
        e: ta.Any = None

        while True:
            try:
                if e is not None:
                    o = g.throw(e)
                else:
                    o = g.send(i)
            except StopIteration as ex:
                return ex.value

            i = None
            e = None

            mo = check.isinstance(o, _MaysyncOp)
            try:
                i = mo.a(mo.mx.s)
            except BaseException as ex:  # noqa
                e = ex

    async def a(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        g = self.mg(*args, **kwargs)

        i: ta.Any = None
        e: ta.Any = None

        while True:
            try:
                if e is not None:
                    o = g.throw(e)
                else:
                    o = g.send(i)
            except StopIteration as ex:
                return ex.value

            i = None
            e = None

            mo = check.isinstance(o, _MaysyncOp)
            try:
                i = await mo.a(mo.mx.a)
            except BaseException as ex:  # noqa
                e = ex


def maysync(m: ta.Callable[..., ta.Awaitable[T]]) -> Maysyncable[T]:
    @functools.wraps(m)
    def mg_fn(*args, **kwargs):
        a = m(*args, **kwargs).__await__()

        try:
            g = iter(a)
            try:
                while True:
                    try:
                        o = g.send(None)
                    except StopIteration as e:
                        return e.value

                    f = check.isinstance(o, _MaysyncFuture)
                    if not f.done:
                        try:
                            f.result = yield f.op
                        except BaseException as e:  # noqa
                            f.error = e
                        f.done = True

            finally:
                g.close()

        finally:
            a.close()

    return _MgMaysyncable(mg_fn)


##


@dc.dataclass(frozen=True, eq=False)
class _MaysyncOp:
    mx: Maysyncable
    a: Args

    def __init_subclass__(cls, **kwargs):
        raise TypeError


##


@dc.dataclass(eq=False)
class _MaysyncFuture(ta.Generic[T]):
    op: _MaysyncOp

    done: bool = False
    error: ta.Optional[BaseException] = None
    result: ta.Optional[T] = None

    def __await__(self):
        if not self.done:
            yield self
        if not self.done:
            raise RuntimeError("await wasn't used with event future")
        if self.error is not None:
            raise self.error
        return self.result



import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh
from ... import reflect as rfl
from ...funcs import match as mfs
from .. import jsonschema as jsch
from .openapi import Reference
from .openapi import Schema


##


def _reference_union_arg(rty: rfl.Type) -> rfl.Type | None:
    if isinstance(rty, rfl.Union) and len(rty.args) == 2 and Reference in rty.args:
        return check.single(a for a in rty.args if a is not Reference)
    else:
        return None


#


@dc.dataclass(frozen=True)
class _ReferenceUnionMarshaler(msh.Marshaler):
    m: msh.Marshaler
    r: msh.Marshaler

    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        if isinstance(o, Reference):
            return self.r.marshal(ctx, o)
        else:
            return self.m.marshal(ctx, o)


class _ReferenceUnionMarshalerFactory(msh.MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: _reference_union_arg(rty) is not None)
    def _build(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        return _ReferenceUnionMarshaler(ctx.make(check.not_none(_reference_union_arg(rty))), ctx.make(Reference))


#


@dc.dataclass(frozen=True)
class _ReferenceUnionUnmarshaler(msh.Unmarshaler):
    u: msh.Unmarshaler
    r: msh.Unmarshaler

    def unmarshal(self, ctx: msh.UnmarshalContext, v: msh.Value) -> ta.Any:
        if not isinstance(v, ta.Mapping):
            raise TypeError(v)
        elif '$ref' in v:
            return self.r.unmarshal(ctx, v)
        else:
            return self.u.unmarshal(ctx, v)  # noqa


class _ReferenceUnionUnmarshalerFactory(msh.UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: _reference_union_arg(rty) is not None)
    def _build(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        return _ReferenceUnionUnmarshaler(ctx.make(check.not_none(_reference_union_arg(rty))), ctx.make(Reference))


##


@dc.dataclass(frozen=True)
class _SchemaMarshaler(msh.Marshaler):
    m_dct: ta.Mapping[str, tuple[str, msh.Marshaler]]
    kw_m: msh.Marshaler

    def marshal(self, ctx: msh.MarshalContext, o: ta.Any) -> msh.Value:
        sch: Schema = check.isinstance(o, Schema)
        dct = {}
        for f, (k, fm) in self.m_dct.items():
            fv = getattr(sch, f)
            if fv is None:
                continue
            dct[k] = fm.marshal(ctx, fv)
        if sch.keywords is not None:
            dct.update(keywords=self.kw_m.marshal(ctx, sch.keywords))
        return dct


class _SchemaMarshalerFactory(msh.MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: rty is Schema)
    def _build(self, ctx: msh.MarshalContext, rty: rfl.Type) -> msh.Marshaler:
        return _SchemaMarshaler(
            {
                f: (msh.translate_name(f, msh.Naming.LOW_CAMEL), ctx.make(rfl.type_(a)))
                for f, a in dc.reflect(Schema).field_annotations.items()
                if f != 'keywords'
            },
            ctx.make(jsch.Keywords),
        )


#


@dc.dataclass(frozen=True)
class _SchemaUnmarshaler(msh.Unmarshaler):
    u_dct: ta.Mapping[str, tuple[str, msh.Unmarshaler]]
    kw_u: msh.Unmarshaler

    def unmarshal(self, ctx: msh.UnmarshalContext, v: msh.Value) -> ta.Any:
        dct = dict(check.isinstance(v, ta.Mapping))
        kw = {}
        for k, (f, fu) in self.u_dct.items():
            try:
                kv = dct.pop(k)
            except KeyError:
                continue
            kw[f] = fu.unmarshal(ctx, kv)
        if dct:
            kw.update(keywords=self.kw_u.unmarshal(ctx, dct))
        return Schema(**kw)


class _SchemaUnmarshalerFactory(msh.UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: rty is Schema)
    def _build(self, ctx: msh.UnmarshalContext, rty: rfl.Type) -> msh.Unmarshaler:
        return _SchemaUnmarshaler(
            {
                msh.translate_name(f, msh.Naming.LOW_CAMEL): (f, ctx.make(rfl.type_(a)))
                for f, a in dc.reflect(Schema).field_annotations.items()
                if f != 'keywords'
            },
            ctx.make(jsch.Keywords),
        )


##


@lang.static_init
def _install_standard_marshalling() -> None:
    msh.install_standard_factories(
        _ReferenceUnionMarshalerFactory(),
        _ReferenceUnionUnmarshalerFactory(),

        _SchemaMarshalerFactory(),
        _SchemaUnmarshalerFactory(),
    )



# ruff: noqa: UP007 UP045
# @omlish-lite
# Based on bluelet ( https://github.com/sampsyo/bluelet ) by Adrian Sampson, original license:
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import abc
import dataclasses as dc
import typing as ta


R = ta.TypeVar('R')

BlueletEventT = ta.TypeVar('BlueletEventT', bound='BlueletEvent')

BlueletWaitable = ta.Union[int, 'BlueletHasFileno']  # ta.TypeAlias


##


class BlueletEvent(abc.ABC):  # noqa
    """
    Just a base class identifying Bluelet events. An event is an object yielded from a Bluelet coro coroutine to
    suspend operation and communicate with the scheduler.
    """

    def __await__(self):
        return BlueletFuture(self).__await__()


##


class BlueletHasFileno(ta.Protocol):
    def fileno(self) -> int: ...


@dc.dataclass(frozen=True)
class BlueletWaitables:
    r: ta.Sequence[BlueletWaitable] = ()
    w: ta.Sequence[BlueletWaitable] = ()
    x: ta.Sequence[BlueletWaitable] = ()


class WaitableBlueletEvent(BlueletEvent, abc.ABC):  # noqa
    """
    A waitable event is one encapsulating an action that can be waited for using a select() call. That is, it's an event
    with an associated file descriptor.
    """

    def waitables(self) -> BlueletWaitables:
        """
        Return "waitable" objects to pass to select(). Should return three iterables for input readiness, output
        readiness, and exceptional conditions (i.e., the three lists passed to select()).
        """
        return BlueletWaitables()

    def fire(self) -> ta.Any:
        """Called when an associated file descriptor becomes ready (i.e., is returned from a select() call)."""


##


@dc.dataclass(eq=False)
class BlueletFuture(ta.Generic[BlueletEventT, R]):
    event: BlueletEventT
    done: bool = False
    result: ta.Optional[R] = None

    def __await__(self):
        if not self.done:
            yield self
        if not self.done:
            raise RuntimeError("await wasn't used with event future")
        return self.result



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import contextlib
import logging
import os
import subprocess
import time
import typing as ta

from ..lite.timeouts import Timeout
from .wrap import subprocess_maybe_shell_wrap_exec


T = ta.TypeVar('T')
SubprocessChannelOption = ta.Literal['pipe', 'stdout', 'devnull']  # ta.TypeAlias


##


# Valid channel type kwarg values:
#  - A special flag negative int
#  - A positive fd int
#  - A file-like object
#  - None

SUBPROCESS_CHANNEL_OPTION_VALUES: ta.Mapping[SubprocessChannelOption, int] = {
    'pipe': subprocess.PIPE,
    'stdout': subprocess.STDOUT,
    'devnull': subprocess.DEVNULL,
}


##


class VerboseCalledProcessError(subprocess.CalledProcessError):
    @classmethod
    def from_std(cls, e: subprocess.CalledProcessError) -> 'VerboseCalledProcessError':
        return cls(
            e.returncode,
            e.cmd,
            output=e.output,
            stderr=e.stderr,
        )

    def __str__(self) -> str:
        msg = super().__str__()
        if self.output is not None:
            msg += f' Output: {self.output!r}'
        if self.stderr is not None:
            msg += f' Stderr: {self.stderr!r}'
        return msg


class BaseSubprocesses(abc.ABC):  # noqa
    DEFAULT_LOGGER: ta.ClassVar[ta.Optional[logging.Logger]] = None

    def __init__(
            self,
            *,
            log: ta.Optional[logging.Logger] = None,
            try_exceptions: ta.Optional[ta.Tuple[ta.Type[Exception], ...]] = None,
    ) -> None:
        super().__init__()

        self._log = log if log is not None else self.DEFAULT_LOGGER
        self._try_exceptions = try_exceptions if try_exceptions is not None else self.DEFAULT_TRY_EXCEPTIONS

    def set_logger(self, log: ta.Optional[logging.Logger]) -> None:
        self._log = log

    #

    def prepare_args(
            self,
            *cmd: str,
            env: ta.Optional[ta.Mapping[str, ta.Any]] = None,
            extra_env: ta.Optional[ta.Mapping[str, ta.Any]] = None,
            quiet: bool = False,
            shell: bool = False,
            **kwargs: ta.Any,
    ) -> ta.Tuple[ta.Tuple[ta.Any, ...], ta.Dict[str, ta.Any]]:
        if self._log:
            self._log.debug('Subprocesses.prepare_args: cmd=%r', cmd)
            if extra_env:
                self._log.debug('Subprocesses.prepare_args: extra_env=%r', extra_env)

        #

        if extra_env:
            env = {**(env if env is not None else os.environ), **extra_env}

        #

        if quiet and 'stderr' not in kwargs:
            if self._log and not self._log.isEnabledFor(logging.DEBUG):
                kwargs['stderr'] = subprocess.DEVNULL

        for chk in ('stdout', 'stderr'):
            try:
                chv = kwargs[chk]
            except KeyError:
                continue
            kwargs[chk] = SUBPROCESS_CHANNEL_OPTION_VALUES.get(chv, chv)

        #

        if not shell:
            cmd = subprocess_maybe_shell_wrap_exec(*cmd)

        #

        if 'timeout' in kwargs:
            kwargs['timeout'] = Timeout.of(kwargs['timeout']).or_(None)

        #

        return cmd, dict(
            env=env,
            shell=shell,
            **kwargs,
        )

    @contextlib.contextmanager
    def wrap_call(
            self,
            *cmd: ta.Any,
            raise_verbose: bool = False,
            **kwargs: ta.Any,
    ) -> ta.Iterator[None]:
        start_time = time.time()
        try:
            if self._log:
                self._log.debug('Subprocesses.wrap_call.try: cmd=%r', cmd)

            yield

        except Exception as exc:  # noqa
            if self._log:
                self._log.debug('Subprocesses.wrap_call.except: exc=%r', exc)

            if (
                    raise_verbose and
                    isinstance(exc, subprocess.CalledProcessError) and
                    not isinstance(exc, VerboseCalledProcessError) and
                    (exc.output is not None or exc.stderr is not None)
            ):
                raise VerboseCalledProcessError.from_std(exc) from exc

            raise

        finally:
            end_time = time.time()
            elapsed_s = end_time - start_time

            if self._log:
                self._log.debug('Subprocesses.wrap_call.finally: elapsed_s=%f cmd=%r', elapsed_s, cmd)

    @contextlib.contextmanager
    def prepare_and_wrap(
            self,
            *cmd: ta.Any,
            raise_verbose: bool = False,
            **kwargs: ta.Any,
    ) -> ta.Iterator[ta.Tuple[
        ta.Tuple[ta.Any, ...],
        ta.Dict[str, ta.Any],
    ]]:
        cmd, kwargs = self.prepare_args(*cmd, **kwargs)

        with self.wrap_call(
                *cmd,
                raise_verbose=raise_verbose,
                **kwargs,
        ):
            yield cmd, kwargs

    #

    DEFAULT_TRY_EXCEPTIONS: ta.Tuple[ta.Type[Exception], ...] = (
        FileNotFoundError,
        subprocess.CalledProcessError,
    )

    def try_fn(
            self,
            fn: ta.Callable[..., T],
            *cmd: str,
            try_exceptions: ta.Optional[ta.Tuple[ta.Type[Exception], ...]] = None,
            **kwargs: ta.Any,
    ) -> ta.Union[T, Exception]:
        if try_exceptions is None:
            try_exceptions = self._try_exceptions

        try:
            return fn(*cmd, **kwargs)

        except try_exceptions as e:  # noqa
            if self._log and self._log.isEnabledFor(logging.DEBUG):
                self._log.exception('command failed')
            return e

    async def async_try_fn(
            self,
            fn: ta.Callable[..., ta.Awaitable[T]],
            *cmd: ta.Any,
            try_exceptions: ta.Optional[ta.Tuple[ta.Type[Exception], ...]] = None,
            **kwargs: ta.Any,
    ) -> ta.Union[T, Exception]:
        if try_exceptions is None:
            try_exceptions = self._try_exceptions

        try:
            return await fn(*cmd, **kwargs)

        except try_exceptions as e:  # noqa
            if self._log and self._log.isEnabledFor(logging.DEBUG):
                self._log.exception('command failed')
            return e



import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from ..coro import BytesSteppedCoro
from .base import Compression
from .base import IncrementalCompression
from .codecs import make_compression_codec
from .codecs import make_compression_lazy_loaded_codec


if ta.TYPE_CHECKING:
    import lz4.frame as lz4_frame
else:
    lz4_frame = lang.proxy_import('lz4.frame')


##


@dc.dataclass(frozen=True, kw_only=True)
class Lz4Compression(Compression, IncrementalCompression):
    level: int = 0

    block_size: int = 0
    block_linked: bool = True
    block_checksum: bool = False
    content_checksum: bool = False
    store_size: bool = True
    auto_flush: bool = False

    def compress(self, d: bytes) -> bytes:
        return lz4_frame.compress(
            d,
            compression_level=self.level,
            block_size=self.block_size,
            content_checksum=self.content_checksum,
            block_linked=self.block_linked,
            store_size=self.store_size,
        )

    def decompress(self, d: bytes) -> bytes:
        return lz4_frame.decompress(
            d,
        )

    @lang.autostart
    def compress_incremental(self) -> BytesSteppedCoro[None]:
        with lz4_frame.LZ4FrameCompressor(
                compression_level=self.level,
                block_size=self.block_size,
                block_linked=self.block_linked,
                block_checksum=self.block_checksum,
                content_checksum=self.content_checksum,
                auto_flush=self.auto_flush,
        ) as compressor:
            started = False
            while True:
                i = check.isinstance((yield None), bytes)
                if not started:
                    yield compressor.begin()
                    started = True
                if not i:
                    yield compressor.flush()
                    yield b''
                    return
                if (o := compressor.compress(i)):
                    yield o

    @lang.autostart
    def decompress_incremental(self) -> BytesSteppedCoro[None]:
        # lz4 lib does internal buffering so this is simply a BytesSteppedCoro not a BytesSteppedReaderCoro as it
        # only yields None, accepting any number of bytes at a time.
        with lz4_frame.LZ4FrameDecompressor() as decompressor:
            while True:
                i = check.isinstance((yield None), bytes)
                if not i:
                    yield b''
                    return
                if (o := decompressor.decompress(i)):
                    yield o


##


LZ4_CODEC = make_compression_codec('lz4', Lz4Compression)

# @omlish-manifest
_LZ4_LAZY_CODEC = make_compression_lazy_loaded_codec(__name__, 'LZ4_CODEC', LZ4_CODEC)



# type: ignore
# ruff: noqa
# flake8: noqa
# @omlish-generated
# Generated from Protobuf3.g4 by ANTLR 4.13.2
from ....text.antlr._runtime._all import *
if "." in __name__:
    from .Protobuf3Parser import Protobuf3Parser
else:
    from Protobuf3Parser import Protobuf3Parser

# This class defines a complete generic visitor for a parse tree produced by Protobuf3Parser.

class Protobuf3Visitor(ParseTreeVisitor):

    # Visit a parse tree produced by Protobuf3Parser#proto.
    def visitProto(self, ctx:Protobuf3Parser.ProtoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#syntax.
    def visitSyntax(self, ctx:Protobuf3Parser.SyntaxContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#syntaxExtra.
    def visitSyntaxExtra(self, ctx:Protobuf3Parser.SyntaxExtraContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#importStmt.
    def visitImportStmt(self, ctx:Protobuf3Parser.ImportStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#packageStmt.
    def visitPackageStmt(self, ctx:Protobuf3Parser.PackageStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#option.
    def visitOption(self, ctx:Protobuf3Parser.OptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#optionName.
    def visitOptionName(self, ctx:Protobuf3Parser.OptionNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#optionBody.
    def visitOptionBody(self, ctx:Protobuf3Parser.OptionBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#optionBodyVariable.
    def visitOptionBodyVariable(self, ctx:Protobuf3Parser.OptionBodyVariableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#topLevelDef.
    def visitTopLevelDef(self, ctx:Protobuf3Parser.TopLevelDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#message.
    def visitMessage(self, ctx:Protobuf3Parser.MessageContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#messageBody.
    def visitMessageBody(self, ctx:Protobuf3Parser.MessageBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#messageBodyContent.
    def visitMessageBodyContent(self, ctx:Protobuf3Parser.MessageBodyContentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#enumDef.
    def visitEnumDef(self, ctx:Protobuf3Parser.EnumDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#enumBody.
    def visitEnumBody(self, ctx:Protobuf3Parser.EnumBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#enumField.
    def visitEnumField(self, ctx:Protobuf3Parser.EnumFieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#enumValueOption.
    def visitEnumValueOption(self, ctx:Protobuf3Parser.EnumValueOptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#extend.
    def visitExtend(self, ctx:Protobuf3Parser.ExtendContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#service.
    def visitService(self, ctx:Protobuf3Parser.ServiceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#rpc.
    def visitRpc(self, ctx:Protobuf3Parser.RpcContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#reserved.
    def visitReserved(self, ctx:Protobuf3Parser.ReservedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#ranges.
    def visitRanges(self, ctx:Protobuf3Parser.RangesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#rangeRule.
    def visitRangeRule(self, ctx:Protobuf3Parser.RangeRuleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#fieldNames.
    def visitFieldNames(self, ctx:Protobuf3Parser.FieldNamesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#typeRule.
    def visitTypeRule(self, ctx:Protobuf3Parser.TypeRuleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#simpleType.
    def visitSimpleType(self, ctx:Protobuf3Parser.SimpleTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#fieldNumber.
    def visitFieldNumber(self, ctx:Protobuf3Parser.FieldNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#field.
    def visitField(self, ctx:Protobuf3Parser.FieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#fieldOptions.
    def visitFieldOptions(self, ctx:Protobuf3Parser.FieldOptionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#fieldOption.
    def visitFieldOption(self, ctx:Protobuf3Parser.FieldOptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#oneof.
    def visitOneof(self, ctx:Protobuf3Parser.OneofContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#oneofField.
    def visitOneofField(self, ctx:Protobuf3Parser.OneofFieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#mapField.
    def visitMapField(self, ctx:Protobuf3Parser.MapFieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#keyType.
    def visitKeyType(self, ctx:Protobuf3Parser.KeyTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#reservedWord.
    def visitReservedWord(self, ctx:Protobuf3Parser.ReservedWordContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#fullIdent.
    def visitFullIdent(self, ctx:Protobuf3Parser.FullIdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#messageName.
    def visitMessageName(self, ctx:Protobuf3Parser.MessageNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#enumName.
    def visitEnumName(self, ctx:Protobuf3Parser.EnumNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#messageOrEnumName.
    def visitMessageOrEnumName(self, ctx:Protobuf3Parser.MessageOrEnumNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#fieldName.
    def visitFieldName(self, ctx:Protobuf3Parser.FieldNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#oneofName.
    def visitOneofName(self, ctx:Protobuf3Parser.OneofNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#mapName.
    def visitMapName(self, ctx:Protobuf3Parser.MapNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#serviceName.
    def visitServiceName(self, ctx:Protobuf3Parser.ServiceNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#rpcName.
    def visitRpcName(self, ctx:Protobuf3Parser.RpcNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#messageType.
    def visitMessageType(self, ctx:Protobuf3Parser.MessageTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#messageOrEnumType.
    def visitMessageOrEnumType(self, ctx:Protobuf3Parser.MessageOrEnumTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#emptyStmt.
    def visitEmptyStmt(self, ctx:Protobuf3Parser.EmptyStmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Protobuf3Parser#constant.
    def visitConstant(self, ctx:Protobuf3Parser.ConstantContext):
        return self.visitChildren(ctx)



del Protobuf3Parser


# ruff: noqa: UP006 UP007 UP045
import types
import typing as ta


##


def import_module(dotted_path: str) -> types.ModuleType:
    if not dotted_path:
        raise ImportError(dotted_path)
    mod = __import__(dotted_path, globals(), locals(), [])
    for name in dotted_path.split('.')[1:]:
        try:
            mod = getattr(mod, name)
        except AttributeError:
            raise AttributeError(f'Module {mod!r} has no attribute {name!r}') from None
    return mod


def import_module_attr(dotted_path: str) -> ta.Any:
    module_name, _, class_name = dotted_path.rpartition('.')
    mod = import_module(module_name)
    try:
        return getattr(mod, class_name)
    except AttributeError:
        raise AttributeError(f'Module {module_name!r} has no attr {class_name!r}') from None


def import_attr(dotted_path: str) -> ta.Any:
    importlib = __import__('importlib')
    parts = dotted_path.split('.')
    mod: ta.Any = None
    mod_pos = 0
    while mod_pos < len(parts):
        mod_name = '.'.join(parts[:mod_pos + 1])
        try:
            mod = importlib.import_module(mod_name)
        except ImportError:
            break
        mod_pos += 1
    if mod is None:
        raise ImportError(dotted_path)
    obj = mod
    for att_pos in range(mod_pos, len(parts)):
        obj = getattr(obj, parts[att_pos])
    return obj



from .... import lang
from .base import AnyArrayKeyword
from .base import AnyKeyword
from .base import BooleanKeyword
from .base import BooleanOrKeywordsKeyword
from .base import KeywordsArrayKeyword
from .base import KeywordsKeyword
from .base import KnownKeyword
from .base import NumberKeyword
from .base import StrOrStrArrayKeyword
from .base import StrToKeywordsKeyword


##


class ValidationKeyword(KnownKeyword, lang.Abstract, lang.Sealed):
    pass


##


class Type(StrOrStrArrayKeyword, ValidationKeyword, lang.Final, tag='type'):
    pass


class Const(AnyKeyword, ValidationKeyword, lang.Final, tag='const'):
    pass


class Enum(AnyArrayKeyword, ValidationKeyword, lang.Final, tag='enum'):
    pass


class Items(KeywordsKeyword, ValidationKeyword, lang.Final, tag='items'):
    pass


class Required(StrOrStrArrayKeyword, ValidationKeyword, lang.Final, tag='required'):
    pass


class Properties(StrToKeywordsKeyword, ValidationKeyword, lang.Final, tag='properties'):
    pass


class AdditionalProperties(BooleanOrKeywordsKeyword, ValidationKeyword, lang.Final, tag='additionalProperties'):
    pass


##


class MaxItems(NumberKeyword, ValidationKeyword, lang.Final, tag='maxItems'):
    pass


class MinItems(NumberKeyword, ValidationKeyword, lang.Final, tag='minItems'):
    pass


class UniqueItems(BooleanKeyword, ValidationKeyword, lang.Final, tag='uniqueItems'):
    pass


#


class Maximum(NumberKeyword, ValidationKeyword, lang.Final, tag='maximum'):
    pass


class ExclusiveMaximum(NumberKeyword, ValidationKeyword, lang.Final, tag='exclusiveMaximum'):
    pass


class Minimum(NumberKeyword, ValidationKeyword, lang.Final, tag='minimum'):
    pass


class ExclusiveMinimum(NumberKeyword, ValidationKeyword, lang.Final, tag='exclusiveMinimum'):
    pass


#


class AnyOf(KeywordsArrayKeyword, ValidationKeyword, lang.Final, tag='anyOf'):
    pass


class OneOf(KeywordsArrayKeyword, ValidationKeyword, lang.Final, tag='oneOf'):
    pass



"""
FIXME:
 - don't base64 by default, only for json-esque targets
"""
import base64
import typing as ta

from ... import check
from ... import dataclasses as dc
from ..base import MarshalContext
from ..base import Marshaler
from ..base import TypeMapMarshalerFactory
from ..base import TypeMapUnmarshalerFactory
from ..base import UnmarshalContext
from ..base import Unmarshaler
from ..values import Value


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True)
class Base64MarshalerUnmarshaler(Marshaler, Unmarshaler, ta.Generic[T]):
    ty: type[T]

    def marshal(self, ctx: MarshalContext, o: bytes) -> str:
        return base64.b64encode(o).decode()

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> T:
        return self.ty(base64.b64decode(check.isinstance(v, str).encode()))  # type: ignore


BASE64_TYPES = (
    bytes,
    bytearray,
)

BASE64_MARSHALER_FACTORY = TypeMapMarshalerFactory({
    ty: Base64MarshalerUnmarshaler(ty) for ty in BASE64_TYPES
})

BASE64_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({
    ty: Base64MarshalerUnmarshaler(ty) for ty in BASE64_TYPES
})



import dataclasses as dc
import typing as ta

import anyio.streams.memory
import anyio.streams.stapled

from ... import check


T = ta.TypeVar('T')


##


MemoryObjectReceiveStream: ta.TypeAlias = anyio.streams.memory.MemoryObjectReceiveStream
MemoryObjectSendStream: ta.TypeAlias = anyio.streams.memory.MemoryObjectSendStream

StapledByteStream: ta.TypeAlias = anyio.streams.stapled.StapledByteStream
StapledObjectStream: ta.TypeAlias = anyio.streams.stapled.StapledObjectStream


@dc.dataclass(eq=False)
class MemoryStapledObjectStream(StapledObjectStream[T]):
    send_stream: MemoryObjectSendStream[T]
    receive_stream: MemoryObjectReceiveStream[T]


def split_memory_object_streams(
        *args: anyio.create_memory_object_stream[T],
) -> tuple[
    MemoryObjectSendStream[T],
    MemoryObjectReceiveStream[T],
]:
    [tup] = args
    return tup  # noqa


def create_stapled_memory_object_stream(max_buffer_size: float = 0) -> MemoryStapledObjectStream:
    return MemoryStapledObjectStream(*anyio.create_memory_object_stream(max_buffer_size))


# FIXME: https://github.com/python/mypy/issues/15238
# FIXME: https://youtrack.jetbrains.com/issues?q=tag:%20%7BPEP%20695%7D
def create_memory_object_stream[T](max_buffer_size: float = 0) -> tuple[
    MemoryObjectSendStream[T],
    MemoryObjectReceiveStream[T],
]:
    return anyio.create_memory_object_stream[T](max_buffer_size)  # noqa


def staple_memory_object_stream(
        *args: anyio.create_memory_object_stream[T],
) -> MemoryStapledObjectStream[T]:
    send, receive = args
    return MemoryStapledObjectStream(
        check.isinstance(send, MemoryObjectSendStream),
        check.isinstance(receive, MemoryObjectReceiveStream),
    )


# FIXME: https://github.com/python/mypy/issues/15238
# FIXME: https://youtrack.jetbrains.com/issues?q=tag:%20%7BPEP%20695%7D
def staple_memory_object_stream2[T](max_buffer_size: float = 0) -> MemoryStapledObjectStream[T]:
    send, receive = anyio.create_memory_object_stream[T](max_buffer_size)
    return MemoryStapledObjectStream(
        check.isinstance(send, MemoryObjectSendStream),
        check.isinstance(receive, MemoryObjectReceiveStream),
    )



import functools
import typing as ta

from ... import dataclasses as dc
from ... import lang
from ..headers import HttpHeaders
from .base import HttpClient
from .base import HttpClientError
from .base import HttpRequest
from .base import StreamHttpResponse


if ta.TYPE_CHECKING:
    import httpx
else:
    httpx = lang.proxy_import('httpx')


##


class HttpxHttpClient(HttpClient):
    @dc.dataclass(frozen=True)
    class _StreamAdapter:
        it: ta.Iterator[bytes]

        def read(self, /, n: int = -1) -> bytes:
            if n < 0:
                return b''.join(self.it)
            else:
                try:
                    return next(self.it)
                except StopIteration:
                    return b''

    def _stream_request(self, req: HttpRequest) -> StreamHttpResponse:
        try:
            resp_cm = httpx.stream(
                method=req.method_or_default,
                url=req.url,
                headers=req.headers_ or None,  # type: ignore
                content=req.data,
                timeout=req.timeout_s,
            )

        except httpx.HTTPError as e:
            raise HttpClientError from e

        resp_close = functools.partial(resp_cm.__exit__, None, None, None)

        try:
            resp = resp_cm.__enter__()
            return StreamHttpResponse(
                status=resp.status_code,
                headers=HttpHeaders(resp.headers.raw),
                request=req,
                underlying=resp,
                stream=self._StreamAdapter(resp.iter_bytes()),
                _closer=resp_close,  # type: ignore
            )

        except httpx.HTTPError as e:
            resp_close()
            raise HttpClientError from e

        except Exception:
            resp_close()
            raise



# @omlish-lite
import abc
import enum
import os
import typing as ta


##


class TermColor(enum.Enum):
    RED = enum.auto()
    GREEN = enum.auto()
    YELLOW = enum.auto()
    BLUE = enum.auto()


class TermColoring(abc.ABC):
    @abc.abstractmethod
    def color(self, c: TermColor, s: str) -> str:
        raise NotImplementedError

    def red(self, s: str, **kwargs: ta.Any) -> str:
        return self.color(TermColor.RED, s, **kwargs)

    def green(self, s: str, **kwargs: ta.Any) -> str:
        return self.color(TermColor.GREEN, s, **kwargs)

    def yellow(self, s: str, **kwargs: ta.Any) -> str:
        return self.color(TermColor.YELLOW, s, **kwargs)

    def blue(self, s: str, **kwargs: ta.Any) -> str:
        return self.color(TermColor.BLUE, s, **kwargs)


class NopTermColoring(TermColoring):
    def color(self, c: TermColor, s: str) -> str:
        return s


class AnsiTermColoring(TermColoring):
    _COLOR_CODES: ta.ClassVar[ta.Mapping[TermColor, str]] = {
        TermColor.RED: '\033[31m',
        TermColor.GREEN: '\033[32m',
        TermColor.YELLOW: '\033[33m',
        TermColor.BLUE: '\033[34m',
    }

    _RESET_CODE = '\033[0m'

    def color(self, c: TermColor, s: str) -> str:
        return f'{self._COLOR_CODES[c]}{s}{self._RESET_CODE}'


def term_coloring(
        *,
        forced: bool = False,
        disabled: bool = False,
        file: ta.Any = None,
) -> TermColoring:
    if not forced:
        if not disabled:
            if file is not None:
                if hasattr(file, 'isatty'):
                    if not file.isatty():
                        disabled = True
                elif isinstance(file, int):
                    if not os.isatty(file):
                        disabled = True
                else:
                    raise TypeError(file)

        if disabled:
            return NopTermColoring()

    return AnsiTermColoring()


##


if __name__ == '__main__':
    def _main() -> None:
        import argparse

        parser = argparse.ArgumentParser()
        parser.add_argument('--force', action='store_true')
        parser.add_argument('--disable', action='store_true')
        args = parser.parse_args()

        import sys

        tc = term_coloring(
            forced=args.force,
            disabled=args.disable,
            file=sys.stdout,
        )

        print('\n'.join([
            'normal',
            tc.red('red'),
            tc.green('green'),
            tc.yellow('yellow'),
            tc.blue('blue'),
            'normal',
        ]))

    _main()



import typing as ta

from .. import dataclasses as dc
from .content import Content
from .content import Dom
from .content import kwargs_to_attrs


##


def d(
        tag: str,
        *attrs_and_contents: tuple[str, ta.Any] | Content,
        **kwargs: ta.Any,
) -> Dom:
    c = []
    for a in attrs_and_contents:
        if isinstance(a, tuple):
            k, v = a
            if k in kwargs:
                raise KeyError(f'Attribute {k} already set')
            kwargs[k] = v
        else:
            c.append(a)

    return Dom(
        tag,
        attrs=kwargs_to_attrs(**kwargs) or None,
        body=c or None,
    )


##


@dc.dataclass(frozen=True)
class DomBuilder:
    tag: str

    def __call__(
            self,
            *attrs_and_contents: tuple[str, ta.Any] | Content,
            **kwargs: ta.Any,
    ) -> Dom:
        return d(self.tag, *attrs_and_contents, **kwargs)


class DomAccessor:
    def __getattr__(self, tag: str) -> DomBuilder:
        return DomBuilder(tag)


D = DomAccessor()



# ruff: noqa: UP006 UP007 UP045
"""
https://docs.python.org/3/library/unittest.html#command-line-interface
~ https://github.com/python/cpython/tree/f66c75f11d3aeeb614600251fd5d3fe1a34b5ff1/Lib/unittest
"""
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001-2024 Python Software Foundation; All Rights
# Reserved" are retained in Python alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import contextlib
import dataclasses as dc
import sys
import time
import typing as ta
import unittest
import warnings

from .types import UnittestTest


##


class UnittestTestRunner:
    """
    A test runner class that displays results in textual form.

    It prints out the names of tests as they are run, errors as they occur, and a summary of the results at the end of
    the test run.
    """

    @dc.dataclass(frozen=True)
    class Args:
        descriptions: bool = True
        verbosity: int = 1
        failfast: bool = False
        buffer: bool = False
        warnings: ta.Optional[str] = None
        tb_locals: bool = False
        catchbreak: bool = False

    def __init__(
            self,
            args: Args = Args(),
            *,
            stream: ta.Optional[ta.Any] = None,
    ):
        super().__init__()

        self._args = args

        if stream is None:
            stream = sys.stderr
        self._stream = UnittestTestRunner._WritelnDecorator(stream)

    #

    class _WritelnDecorator:
        def __init__(self, stream):
            super().__init__()

            self.stream = stream

        def __getattr__(self, attr):
            if attr in ('stream', '__getstate__'):
                raise AttributeError(attr)
            return getattr(self.stream, attr)

        def writeln(self, arg=None):
            if arg:
                self.write(arg)
            self.write('\n')  # text-mode streams translate to \r\n if needed

    #

    @contextlib.contextmanager
    def _warnings_context(self) -> ta.Iterator[None]:
        with warnings.catch_warnings():
            w = self._args.warnings

            if w is None and not sys.warnoptions:
                # Even if DeprecationWarnings are ignored by default print them anyway unless other warnings settings
                # are specified by the warnings arg or the -W python flag.
                w = 'default'  # noqa
            else:
                # Here self.warnings is set either to the value passed to the warnings args or to None. If the user
                # didn't pass a value self.warnings will be None. This means that the behavior is unchanged and depends
                # on the values passed to -W.
                w = self._args.warnings  # noqa

            if w:
                # If self.warnings is set, use it to filter all the warnings.
                warnings.simplefilter(w)  # noqa

                # If the filter is 'default' or 'always', special-case the warnings from the deprecated unittest methods
                # to show them no more than once per module, because they can be fairly noisy.  The -Wd and -Wa flags
                # can be used to bypass this only when self.warnings is None.
                if w in ['default', 'always']:
                    warnings.filterwarnings(
                        'module',
                        category=DeprecationWarning,
                        message=r'Please use assert\w+ instead.',
                    )

            yield

    #

    def _make_result(self) -> unittest.TextTestResult:
        return unittest.TextTestResult(
            self._stream,  # type: ignore[arg-type]
            self._args.descriptions,
            self._args.verbosity,
        )

    #

    class _InternalRunTestResult(ta.NamedTuple):
        result: unittest.TextTestResult
        time_taken: float

    def _internal_run_test(self, test: ta.Callable[[unittest.TestResult], None]) -> _InternalRunTestResult:
        result = self._make_result()
        unittest.registerResult(result)
        result.failfast = self._args.failfast
        result.buffer = self._args.buffer
        result.tb_locals = self._args.tb_locals

        #

        if self._args.catchbreak:
            unittest.signals.installHandler()

        with self._warnings_context():
            start_time = time.perf_counter()

            start_test_run = getattr(result, 'startTestRun', None)
            if start_test_run is not None:
                start_test_run()

            try:
                test(result)

            finally:
                stop_test_run = getattr(result, 'stopTestRun', None)
                if stop_test_run is not None:
                    stop_test_run()

            stop_time = time.perf_counter()

        time_taken = stop_time - start_time

        return UnittestTestRunner._InternalRunTestResult(
            result,
            time_taken,
        )

    #

    @dc.dataclass(frozen=True)
    class RunResult:
        raw_results: ta.Sequence[unittest.TextTestResult]
        time_taken: float

        num_tests_run: int
        was_successful: bool

        class TestAndReason(ta.NamedTuple):
            test: str
            reason: str

        skipped: ta.Sequence[TestAndReason]
        errors: ta.Sequence[TestAndReason]
        failures: ta.Sequence[TestAndReason]

        expected_failures: ta.Sequence[TestAndReason]
        unexpected_successes: ta.Sequence[str]

        @classmethod
        def merge(cls, results: ta.Iterable['UnittestTestRunner.RunResult']) -> 'UnittestTestRunner.RunResult':
            def reduce_attr(fn, a):
                return fn(getattr(r, a) for r in results)

            def merge_list_attr(a):
                return [rr for r in results for rr in getattr(r, a)]

            return cls(
                raw_results=merge_list_attr('raw_results'),
                time_taken=reduce_attr(sum, 'time_taken'),

                num_tests_run=reduce_attr(sum, 'num_tests_run'),
                was_successful=reduce_attr(all, 'was_successful'),

                skipped=merge_list_attr('skipped'),
                errors=merge_list_attr('errors'),
                failures=merge_list_attr('failures'),

                expected_failures=merge_list_attr('expected_failures'),
                unexpected_successes=merge_list_attr('unexpected_successes'),
            )

    def _build_run_result(self, internal_result: _InternalRunTestResult) -> RunResult:
        result = internal_result.result

        def as_test_and_reasons(l):
            return [
                UnittestTestRunner.RunResult.TestAndReason(result.getDescription(t), r)
                for t, r in l
            ]

        return UnittestTestRunner.RunResult(
            raw_results=[result],
            time_taken=internal_result.time_taken,

            num_tests_run=result.testsRun,
            was_successful=result.wasSuccessful(),

            skipped=as_test_and_reasons(result.skipped),
            errors=as_test_and_reasons(result.errors),
            failures=as_test_and_reasons(result.failures),

            expected_failures=as_test_and_reasons(result.expectedFailures),
            unexpected_successes=[result.getDescription(t) for t in result.unexpectedSuccesses],
        )

    #

    def run(self, test: UnittestTest) -> RunResult:
        return self._build_run_result(self._internal_run_test(test))

    def run_many(self, tests: ta.Iterable[UnittestTest]) -> RunResult:
        return UnittestTestRunner.RunResult.merge([self.run(t) for t in tests])

    #

    separator1 = unittest.TextTestResult.separator1
    separator2 = unittest.TextTestResult.separator2

    def print(self, result: RunResult) -> None:
        if self._args.verbosity > 0:
            self._stream.writeln()
            self._stream.flush()

        for t, r in result.errors:
            self._stream.writeln(self.separator1)
            self._stream.writeln(f'ERROR: {t}')
            self._stream.writeln(self.separator2)
            self._stream.writeln(f'{r}')
            self._stream.flush()

        for t, r in result.failures:
            self._stream.writeln(self.separator1)
            self._stream.writeln(f'FAIL: {t}')
            self._stream.writeln(self.separator2)
            self._stream.writeln(f'{r}')
            self._stream.flush()

        if result.unexpected_successes:
            self._stream.writeln(self.separator1)
            for t in result.unexpected_successes:
                self._stream.writeln(f'UNEXPECTED SUCCESS: {t}')
            self._stream.flush()

        self._stream.writeln(self.separator2)

        self._stream.writeln(
            f'Ran {result.num_tests_run:d} '
            f'test{"s" if result.num_tests_run != 1 else ""} '
            f'in {result.time_taken:.3f}s',
        )
        self._stream.writeln()

        expected_fails = len(result.expected_failures)
        unexpected_successes = len(result.unexpected_successes)
        skipped = len(result.skipped)

        infos: ta.List[str] = []

        if not result.was_successful:
            self._stream.write('FAILED')
            failed, errored = len(result.failures), len(result.errors)
            if failed:
                infos.append(f'failures={failed:d}')
            if errored:
                infos.append(f'errors={errored:d}')
        else:
            self._stream.write('OK')

        if skipped:
            infos.append(f'skipped={skipped:d}')

        if expected_fails:
            infos.append(f'expected failures={expected_fails:d}')

        if unexpected_successes:
            infos.append(f'unexpected successes={unexpected_successes:d}')

        if infos:
            self._stream.writeln(f' ({", ".join(infos)})')
        else:
            self._stream.write('\n')



import enum
import typing as ta


##


class JsonType(enum.Enum):
    NULL = enum.auto()
    BOOLEAN = enum.auto()
    OBJECT = enum.auto()
    ARRAY = enum.auto()
    NUMBER = enum.auto()
    STRING = enum.auto()


TYPE_SETS_BY_JSON_TYPE: ta.Mapping[JsonType, ta.AbstractSet[type]] = {
    JsonType.NULL: {type(None)},
    JsonType.BOOLEAN: {bool},
    JsonType.OBJECT: {dict},
    JsonType.ARRAY: {list},
    JsonType.NUMBER: {int, float},
    JsonType.STRING: {str},
}



# ruff: noqa: N802 UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - structured
 - prefixed
 - debug
 - optional noisy? noisy will never be lite - some kinda configure_standard callback mechanism?
"""
import contextlib
import datetime
import logging
import typing as ta

from .filters import TidLogFilter
from .json import JsonLogFormatter
from .proxy import ProxyLogHandler


##


STANDARD_LOG_FORMAT_PARTS = [
    ('asctime', '%(asctime)-15s'),
    ('process', 'pid=%(process)s'),
    ('thread', 'tid=%(thread)x'),
    ('levelname', '%(levelname)s'),
    ('name', '%(name)s'),
    ('separator', '::'),
    ('message', '%(message)s'),
]


class StandardLogFormatter(logging.Formatter):
    @staticmethod
    def build_log_format(parts: ta.Iterable[ta.Tuple[str, str]]) -> str:
        return ' '.join(v for k, v in parts)

    converter = datetime.datetime.fromtimestamp  # type: ignore

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            return ct.strftime(datefmt)  # noqa
        else:
            t = ct.strftime('%Y-%m-%d %H:%M:%S')
            return '%s.%03d' % (t, record.msecs)  # noqa


##


class StandardConfiguredLogHandler(ProxyLogHandler):
    def __init_subclass__(cls, **kwargs):
        raise TypeError('This class serves only as a marker and should not be subclassed.')


##


@contextlib.contextmanager
def _locking_logging_module_lock() -> ta.Iterator[None]:
    if hasattr(logging, '_acquireLock'):
        logging._acquireLock()  # noqa
        try:
            yield
        finally:
            logging._releaseLock()  # type: ignore  # noqa

    elif hasattr(logging, '_lock'):
        # https://github.com/python/cpython/commit/74723e11109a320e628898817ab449b3dad9ee96
        with logging._lock:  # noqa
            yield

    else:
        raise Exception("Can't find lock in logging module")


def configure_standard_logging(
        level: ta.Union[int, str] = logging.INFO,
        *,
        json: bool = False,
        target: ta.Optional[logging.Logger] = None,
        force: bool = False,
        handler_factory: ta.Optional[ta.Callable[[], logging.Handler]] = None,
) -> ta.Optional[StandardConfiguredLogHandler]:
    with _locking_logging_module_lock():
        if target is None:
            target = logging.root

        #

        if not force:
            if any(isinstance(h, StandardConfiguredLogHandler) for h in list(target.handlers)):
                return None

        #

        if handler_factory is not None:
            handler = handler_factory()
        else:
            handler = logging.StreamHandler()

        #

        formatter: logging.Formatter
        if json:
            formatter = JsonLogFormatter()
        else:
            formatter = StandardLogFormatter(StandardLogFormatter.build_log_format(STANDARD_LOG_FORMAT_PARTS))
        handler.setFormatter(formatter)

        #

        handler.addFilter(TidLogFilter())

        #

        target.addHandler(handler)

        #

        if level is not None:
            target.setLevel(level)

        #

        return StandardConfiguredLogHandler(handler)



# @omlish-lite
# ruff: noqa: UP006 UP007 UP043 UP045
import dataclasses as dc
import enum
import typing as ta


##


class CoroHttpClientErrors:
    class ClientError(Exception):
        pass

    class NotConnectedError(ClientError):
        pass

    class InvalidUrlError(ClientError):
        pass

    @dc.dataclass()
    class UnknownProtocolError(ClientError):
        version: str

    @dc.dataclass()
    class IncompleteReadError(ClientError):
        partial: bytes
        expected: ta.Optional[int] = None

    class ImproperConnectionStateError(ClientError):
        pass

    class CannotSendRequestError(ImproperConnectionStateError):
        pass

    class CannotSendHeaderError(ImproperConnectionStateError):
        pass

    class ResponseNotReadyError(ImproperConnectionStateError):
        pass

    @dc.dataclass()
    class BadStatusLineError(ClientError):
        line: str

    class RemoteDisconnectedError(BadStatusLineError, ConnectionResetError):
        pass

    @dc.dataclass()
    class LineTooLongError(ClientError):
        class LineType(enum.Enum):
            STATUS = enum.auto()
            HEADER = enum.auto()
            CHUNK_SIZE = enum.auto()
            TRAILER = enum.auto()

        line_type: LineType



import abc
import base64
import hashlib
import hmac
import json
import typing as ta

from .. import lang


if ta.TYPE_CHECKING:
    import cryptography.hazmat.primitives.asymmetric.padding
    import cryptography.hazmat.primitives.hashes
    import cryptography.hazmat.primitives.serialization
else:
    cryptography = lang.proxy_import('cryptography', extras=[
        'hazmat.primitives.asymmetric.padding',
        'hazmat.primitives.hashes',
        'hazmat.primitives.serialization',
    ])


##


def as_bytes(v: str | bytes, encoding: str = 'utf-8') -> bytes:
    return v.encode(encoding) if isinstance(v, str) else v


def base64url_encode(b: bytes) -> bytes:
    return base64.urlsafe_b64encode(b).replace(b'=', b'')


##


class Algorithm(abc.ABC):
    @property
    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def prepare_key(self, key: str | bytes) -> ta.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def sign(self, msg: str | bytes, key: ta.Any) -> bytes:
        raise NotImplementedError


class HmacAlgorithm(Algorithm):
    def __init__(self, name: str, digest: ta.Any) -> None:
        super().__init__()
        self._name = name
        self._digest = digest

    @property
    def name(self) -> str:
        return self._name

    def prepare_key(self, key: str | bytes) -> ta.Any:
        return as_bytes(key)

    def sign(self, msg: str | bytes, key: ta.Any) -> bytes:
        return hmac.new(key, as_bytes(msg), self._digest).digest()


class RsaAlgorithm(Algorithm):
    def __init__(self, name: str, digest: str) -> None:
        super().__init__()
        self._name = name
        self._digest = digest

    @property
    def name(self) -> str:
        return self._name

    def prepare_key(self, key: str | bytes) -> ta.Any:
        if (key_bytes := as_bytes(key)).startswith(b'ssh-rsa'):
            return cryptography.hazmat.primitives.serialization.load_ssh_public_key(key_bytes)
        else:
            return cryptography.hazmat.primitives.serialization.load_pem_private_key(key_bytes, password=None)

    def sign(self, msg: str | bytes, key: ta.Any) -> bytes:
        return key.sign(
            as_bytes(msg),
            cryptography.hazmat.primitives.asymmetric.padding.PKCS1v15(),
            getattr(cryptography.hazmat.primitives.hashes, self._digest)(),
        )


ALGORITHMS_BY_NAME = {
    a.name: a for a in [
        HmacAlgorithm('HS256', hashlib.sha256),
        HmacAlgorithm('HS384', hashlib.sha384),
        HmacAlgorithm('HS512', hashlib.sha512),

        RsaAlgorithm('RS256', 'SHA256'),
        RsaAlgorithm('RS384', 'SHA384'),
        RsaAlgorithm('RS512', 'SHA512'),
    ]
}


##


def jwt_encode(
        payload: ta.Mapping[str, ta.Any],
        key: str | bytes,
        *,
        algorithm: str = 'HS256',
) -> str:
    alg = ALGORITHMS_BY_NAME[algorithm]

    segments: list[bytes] = []

    header: dict[str, ta.Any] = {
        'typ': 'jwt',
        'alg': alg.name,
    }
    json_header = json.dumps(header, separators=(',', ':'), sort_keys=True).encode('utf-8')
    segments.append(base64url_encode(json_header))

    json_payload = json.dumps(payload, separators=(',', ':')).encode('utf-8')
    msg_payload = base64url_encode(json_payload)
    segments.append(msg_payload)

    signing_input = b'.'.join(segments)

    key = alg.prepare_key(key)
    signature = alg.sign(signing_input, key)

    segments.append(base64url_encode(signature))

    encoded_string = b'.'.join(segments)
    return encoded_string.decode('utf-8')


def generate_jwt(
        *,
        issuer: str,
        subject: str,
        audience: str,
        issued_at: int,
        expires_at: int,
        scope: str,
        key: str | bytes,
        **kwargs: ta.Any,
) -> str:
    payload = {
        'iss': issuer,
        'sub': subject,
        'aud': audience,
        'iat': issued_at,
        'exp': expires_at,
        'scope': scope,
    }

    return jwt_encode(
        payload,
        key,
        **kwargs,
    )


##


SERVICE_APPLICATION_GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:jwt-bearer'


def build_get_token_body(
        signed_jwt: str,
        *,
        grant_type: str = SERVICE_APPLICATION_GRANT_TYPE,
) -> str:
    return f'grant_type={grant_type}&assertion={signed_jwt}'



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import logging
import threading


##


class TidLogFilter(logging.Filter):
    def filter(self, record):
        # FIXME: handle better - missing from wasm and cosmos
        if hasattr(threading, 'get_native_id'):
            record.tid = threading.get_native_id()
        else:
            record.tid = '?'
        return True



"""
TODO:
 - dialect lol
 - args lol
"""
from ..queries import Stmt
from ..queries.rendering import render
from .asquery import AsQueryParams
from .asquery import as_query_
from .queries import Query
from .queries import QueryMode


##


@as_query_.register
def _(
        stmt: Stmt,
        *,
        params: AsQueryParams,
) -> Query:
    rq = render(stmt)

    if rq.args:
        raise NotImplementedError

    return Query(
        mode=QueryMode.of(params.mode, QueryMode.QUERY),
        text=rq.s,
        args=[],
    )



# Based on pytest-trio, licensed under the MIT license, duplicated below.
#
#  https://github.com/python-trio/pytest-trio/tree/cd6cc14b061d34f35980e38c44052108ed5402d1
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import functools
import sys
import typing as ta

from _pytest.outcomes import Skipped  # noqa
from _pytest.outcomes import XFailed  # noqa

from ...... import cached
from ...... import lang
from ......diag import pydevd as pdu
from .base import AsyncsBackend


if ta.TYPE_CHECKING:
    import trio_asyncio
else:
    trio = lang.proxy_import('trio', extras=['abc'])
    trio_asyncio = lang.proxy_import('trio_asyncio')


##


class TrioAsyncioAsyncsBackend(AsyncsBackend):
    name = 'trio_asyncio'

    def is_available(self) -> bool:
        return lang.can_import('trio_asyncio')

    def is_imported(self) -> bool:
        return 'trio_asyncio' in sys.modules

    #

    @cached.function
    def _prepare(self) -> None:
        # NOTE: Importing it here is apparently necessary to get its patching working - otherwise fails later with
        # `no running event loop` in anyio._backends._asyncio and such.
        import trio_asyncio  # noqa

        if pdu.is_present():
            pdu.patch_for_trio_asyncio()

    def prepare_for_metafunc(self, metafunc) -> None:
        self._prepare()

    #

    def wrap_runner(self, fn):
        @functools.wraps(fn)
        def wrapper(**kwargs):
            return trio_asyncio.run(
                trio_asyncio.aio_as_trio(
                    functools.partial(fn, **kwargs),
                ),
            )

        return wrapper

    async def install_context(self, contextvars_ctx):
        # Seemingly no longer necessary?
        # https://github.com/python-trio/pytest-trio/commit/ef0cd267ea62188a8e475c66cb584e7a2addc02a

        # # This is a gross hack. I guess Trio should provide a context= argument to start_soon/start?
        # task = trio.lowlevel.current_task()
        # if CANARY in task.context:
        #     return

        # task.context = contextvars_ctx

        # # Force a yield so we pick up the new context
        # await trio.sleep(0)

        pass
